#  Copyright (c) 2024 zfit

from __future__ import annotations

import warnings
from typing import TYPE_CHECKING

import jacobi

if TYPE_CHECKING:
    import zfit

from collections.abc import Callable

import logging
from functools import lru_cache, wraps

import numpy as np
import scipy.stats
import tensorflow as tf
from scipy import optimize

import zfit.z.numpy as znp

from .. import z
from ..core.interfaces import ZfitIndependentParameter
from ..core.parameter import assign_values, assign_values_jit
from ..util.container import convert_to_container
from ..util.deprecation import deprecated_args


class NewMinimum(Exception):
    """Exception class for cases where a new minimum is found."""

    pass


class FailEvalLossNaN(Exception):
    pass


class RootFound(Exception):
    """Exception class for cases where a root is found, since SciPy root solvers don't really respect tol or xtol on
    initial evaluation."""

    pass


@deprecated_args(None, "Use cl for confidence level instead.", "sigma")
def compute_errors(
    result: zfit.result.FitResult,
    params: list[ZfitIndependentParameter],
    cl: float | None = None,
    rtol: float | None = None,
    method: str | None = None,
    covariance_method: str | Callable | None = None,
    sigma: float = None,
) -> tuple[
    dict[ZfitIndependentParameter, dict[str, float]],
    zfit.result.FitResult | None,
]:
    """Compute asymmetric errors of parameters by profiling the loss function in the fit result.

    This method finds the value for a given parameter where the loss function is ``cl`` away: for example
    for a cl of 68.3%, this is one (multiplied by the errordef). The other parameters are also minimized and
    not fixed. This method is comparably computationally intensive and, if possible, ``hesse`` should be used.
    However, since ``hesse`` does not capture asymmetric or non-parabolic shaped profiles well, this method is
    preferable.

    Args:
        result: fit result to be used to compute the uncertainties.
        params: The parameters to calculate the
            error. If None, use all parameters.
        cl: Confidence Level of the parameter to be determined. Defaults to 68.3%.
        rtol: relative tol between the computed and the exact roots
        method: type of solver, ``method`` argument of :py:func:`scipy.optimize.root`. Defaults to "hybr".
        covariance_method: The method to use to calculate the correlation matrix, will be forwarded directly
            to :py:meth:`FitResult.covariance`. Valid choices are
            by default {'minuit_hesse', 'hesse_np'} (or any other method defined in the result)
            or a Callable.

    Returns:
        out:
            A ``dict`` containing as keys the parameter and as value a ``dict`` which
            contains two keys 'lower' and 'upper', holding the calculated errors.
            Example: result[par1]['upper'] -> the asymmetric upper error of 'par1'
        out: a fit result is returned when a new minimum is found during the loss scan
    """
    if rtol is None:
        rtol = 0.003
    method = "hybr" if method is None else method
    if cl is None:
        if sigma is None:
            sigma = 1

    else:
        if sigma is None:
            sigma = scipy.stats.chi2(1).ppf(cl) ** 0.5
        else:
            raise ValueError("Cannot specify both sigma and cl.")
    # TODO: how to name things, sigma or cl?

    params = convert_to_container(params)
    new_result = None

    all_params = list(result.params.keys())
    loss = result.loss
    errordef = loss.errordef
    fmin = result.fmin
    rtol *= errordef
    minimizer = result.minimizer

    covariance = result.covariance(method=covariance_method, as_dict=True)
    if covariance is None:
        covariance = result.covariance(method="hesse_np", as_dict=True)
    if covariance is None:
        raise RuntimeError(
            "Could not compute covariance matrix. Check if the minimum is valid."
        )
    param_errors = {param: covariance[(param, param)] ** 0.5 for param in all_params}
    param_scale = np.array(list(param_errors.values()))

    ncalls = 0
    loss_min_tol = minimizer.tol * errordef * 2  # 2 is just to be tolerant
    try:
        to_return = {}
        for param in params:
            assign_values(all_params, result)

            logging.info(f"profiling the parameter {param}")
            param_error = param_errors[param]
            param_value = result.params[param]["value"]

            initial_values = {"lower": [], "upper": []}
            direction = {"lower": -1, "upper": 1}

            for ap in all_params:
                ap_value = result.params[ap]["value"]
                error_factor = (
                    covariance[(param, ap)] * (2 * errordef / param_error**2) ** 0.5
                )
                for d in ["lower", "upper"]:
                    step = direction[d] * error_factor * sigma
                    for ntrial in range(50):
                        ap_value_init = ap_value + step
                        if ap_value_init < ap.lower or ap_value_init > ap.upper:
                            step *= 0.8
                        else:
                            break
                    else:
                        raise RuntimeError(
                            f"Could not find a valid initial value for {ap} in {d} direction after {ntrial + 1} trials."
                            f" step tried: {step}. This should not happes, the error probably looks weird. Maybe plot"
                            f" the loss function for different parameter values and check if it looks reasonable."
                        )

                    initial_values[d].append(ap_value_init)

            index_poi = all_params.index(param)  # remember the index
            _ = loss.value_gradient(
                params=all_params, full=False
            )  # to make sure the loss is compiled

            @z.function(wraps="gradient")
            def optimized_loss_gradient(values, index):
                assert isinstance(index, int)
                assign_values(all_params, values)
                loss_value, gradient = loss.value_gradient(
                    params=all_params, full=False
                )
                if isinstance(gradient, (tuple, list)):
                    gradient = znp.asarray(gradient)
                gradient = znp.concatenate(
                    [gradient[:index_poi], gradient[index_poi + 1 :]]
                )
                return loss_value, gradient

            # TODO: improvement, use jacobian?
            root = None
            ntol = 999  # if it's right in the beginning, we think it's fine

            # TODO: should we add a "robust" or similar option to not skip this?
            # or evaluate and then decide ,maybe use krylov as it doesn't do a lot of calls in the beginning, it
            # approximates the jacobian

            def func(values, args):
                nonlocal ncalls, root, ntol

                ncalls += 1
                swap_sign = args

                try:
                    loss_value, gradient = optimized_loss_gradient(values, index_poi)
                except tf.errors.InvalidArgumentError:
                    msg = (
                        f"The evaluation of the errors of {param.name} failed due to too many NaNs"
                        " being produced in the loss and/or its gradient. This is most probably"
                        " caused by negative values returned from the PDF."
                    )
                    loss_value = znp.array(999999.0)
                    gradient = z.random.normal(stddev=0.1, shape=(len(all_params) - 1,))
                    # raise FailEvalLossNaN(msg)
                zeroed_loss = loss_value.numpy() - fmin

                gradient = znp.array(gradient)

                if swap_sign(param):  # mirror at x-axis to remove second zero
                    zeroed_loss = -zeroed_loss
                    gradient = -gradient
                    logging.info("Swapping sign in error calculation 'zfit_error'")

                elif zeroed_loss < -loss_min_tol:
                    assign_values(all_params, values)  # set values to the new minimum
                    raise NewMinimum("A new minimum is found.")

                downward_shift = errordef * sigma**2
                shifted_loss = zeroed_loss - downward_shift

                if abs(shifted_loss) < rtol:
                    if ntol > 3:
                        root = values[index_poi]
                        raise RootFound()
                    ntol += 1
                else:
                    ntol = 0

                return znp.concatenate([[shifted_loss], gradient])

            to_return[param] = {}
            swap_sign = {
                "lower": lambda p: p > param_value,
                "upper": lambda p: p < param_value,
            }
            for d in ["lower", "upper"]:
                try:
                    root_result = optimize.root(
                        fun=func,
                        args=(swap_sign[d],),
                        x0=np.array(initial_values[d]),
                        tol=rtol * 0.1,  # we won't stop like this anyway
                        options={
                            "diag": 1 / param_scale,  # scale factor for variables
                        },
                        method=method,
                    )
                except RootFound:
                    assert root is not None, "Should be changed inside function."
                else:
                    warnings.warn(
                        f"The root finding did not converge below {rtol} but stopped by its own criteria."
                    )
                    root = root_result.x[index_poi]
                to_return[param][d] = root - param_value

        assign_values(all_params, result)

    except NewMinimum as e:
        from .. import settings

        if settings.get_verbosity() >= 5:
            print(e)
        minimizer = result.minimizer
        loss = result.loss
        new_found_fmin = loss.value(full=False)
        new_result = minimizer.minimize(loss=loss)
        if new_result.fmin >= new_found_fmin + loss_min_tol:
            raise RuntimeError(
                "A new minimum was discovered but the minimizer was not able to find this on himself. "
                "This behavior is currently an exception but will most likely change in the future."
            )
        to_return, new_result_ = compute_errors(
            result=new_result, params=params, cl=cl, rtol=rtol, method=method
        )
        if new_result_ is not None:
            new_result = new_result_
    return to_return, new_result


def numerical_pdf_jacobian(func, params):  # TODO: jit?
    params = list(params.values())
    return znp.asarray(jacobi.jacobi(func, params)[0].T)


# @z.function(wraps="autodiff")
def autodiff_pdf_jacobian(func, params):
    params = list(params.values())
    # TODO(WrappedVariable): this is needed if we want to use wrapped Variables
    # params = z.math._extract_tfparams(params)

    # the below fails for some cases (i.e. CB) with an internal error
    # ValueError: Internal error: Tried to take gradients (or similar) of a variable without handle data:
    # Tensor("Placeholder_1:0", shape=(), dtype=resource)

    # we didn't report that yet, it's too hard to reproduce with a minimal example currently.

    # with tf.GradientTape(watch_accessed_variables=False) as t2:
    #     t2.watch(params)
    #     with tf.GradientTape(watch_accessed_variables=False) as t1:
    #         t1.watch(params)
    #         values = func()
    #
    #     grad = t1.gradient(values, params)
    #     grad = tf.convert_to_tensor(grad)
    # jacobian = t2.jacobian(grad, params)

    columns = []

    for p in params:
        vector = np.zeros(len(params))
        vector[params.index(p)] = 1.0
        with tf.autodiff.ForwardAccumulator(params, list(vector)) as acc:
            values = func()
        columns.append(acc.jvp(values))

    jacobian = z.convert_to_tensor(columns)

    return jacobian


def covariance_with_weights(method, result, params):
    """Compute the covariance matrix of the parameters with weights."""
    from .. import run

    run.assert_executing_eagerly()
    loss = result.loss
    model = loss.model
    data = loss.data
    yields = None
    if loss.is_extended:
        yields = []
        for m in model:
            yields.append(m.get_yield())
    constraints = None
    if loss.constraints:
        constraints = loss.constraints

    old_vals = run(params)

    Hinv_dict = method(result=result, params=params)  # inverse of the hessian matrix
    if not Hinv_dict:
        return {}
    Hinv = dict_to_matrix(params, Hinv_dict)

    def func():
        values = []
        for i, (m, d) in enumerate(zip(model, data)):
            v = m.log_pdf(d)
            weights = d.weights
            if weights is not None:
                v *= weights
            values.append(v)
            if yields is not None:
                yi = yields[i]
                nevents_collected = (
                    tf.reduce_sum(weights) if weights is not None else d.nevents
                )
                term_new = tf.nn.log_poisson_loss(
                    nevents_collected, znp.log(yi), compute_full_loss=True
                )[
                    ..., None
                ]  # make it an array like the others
                values.append(term_new)
        if constraints is not None:
            for constraint in constraints:
                values.append(znp.reshape(constraint.value(), (-1,)))

        return znp.concatenate(values, axis=0)

    params_dict = {p.name: p for p in params}

    if run.get_autograd_mode():
        try:
            jacobian = autodiff_pdf_jacobian(func=func, params=params_dict)
        except ValueError as error:
            raise ValueError(
                "The autodiff could not be performed for the jacobian matrix. This can have a natural cause (see error above)"
                " or be a bug in the autodiff. In the latter case, you can try to use the numerical jacobian instead using:\n"
                "with zfit.run.set_autograd_mode(False):\n"
                "    # calculate here, i.e. result.errors()"
            ) from error
    else:

        def wrapped_func(values):
            assign_values(list(params_dict.values()), values)
            return np.array(func())

        jacobian = numerical_pdf_jacobian(func=wrapped_func, params=params_dict)

    C = znp.matmul(jacobian, znp.transpose(jacobian))
    covariance = np.asarray(znp.matmul(Hinv, znp.matmul(C, Hinv)))
    assign_values(params, old_vals)
    return matrix_to_dict(params, covariance)


def dict_to_matrix(params, matrix_dict):
    nparams = len(params)
    matrix = np.empty((nparams, nparams))
    if not matrix_dict:
        return None

    for i in range(nparams):
        pi = params[i]
        for j in range(i, nparams):
            pj = params[j]
            matrix[i, j] = matrix_dict[(pi, pj)]
            if i != j:
                matrix[j, i] = matrix_dict[(pi, pj)]

    return matrix


def matrix_to_dict(params, matrix):
    nparams = len(params)
    matrix_dict = {}
    if matrix is None:
        return matrix_dict

    for i in range(nparams):
        pi = params[i]
        for j in range(i, nparams):
            pj = params[j]
            matrix_dict[(pi, pj)] = matrix[i, j]
            if i != j:
                matrix_dict[(pj, pi)] = matrix[i, j]

    return matrix_dict


def np_cache(*args, **kwargs):
    """LRU cache implementation for functions whose FIRST parameter is a numpy array.

    >>> array = np.array([[1, 2, 3], [4, 5, 6]])
    >>> @np_cache(maxsize=256)
    ... def multiply(array, factor):
    ...     print("Calculating...")
    ...     return factor*array
    >>> multiply(array, 2)
    Calculating...
    array([[ 2,  4,  6],
           [ 8, 10, 12]])
    >>> multiply(array, 2)
    array([[ 2,  4,  6],
           [ 8, 10, 12]])
    >>> multiply.cache_info()
    CacheInfo(hits=1, misses=1, maxsize=256, currsize=1)
    """

    def decorator(function):
        @wraps(function)
        def wrapper(np_array, *args, **kwargs):
            hashable_array = array_to_tuple(np_array)
            return cached_wrapper(hashable_array, *args, **kwargs)

        @lru_cache(*args, **kwargs)
        def cached_wrapper(hashable_array, *args, **kwargs):
            array = np.array(hashable_array)
            return function(array, *args, **kwargs)

        def array_to_tuple(np_array):
            """Iterates recursively."""
            try:
                return tuple(array_to_tuple(_) for _ in np_array)
            except TypeError:
                return np_array

        # copy lru_cache attributes over too
        wrapper.cache_info = cached_wrapper.cache_info
        wrapper.cache_clear = cached_wrapper.cache_clear

        return wrapper

    return decorator



#  Copyright (c) 2023 zfit
import numpy as np
import pytest
import tensorflow as tf
from ordered_set import OrderedSet

import zfit
from zfit import Parameter, z
from zfit.exception import NameAlreadyTakenError, LogicalUndefinedOperationError
from zfit.param import ComplexParameter, ComposedParameter


def test_complex_param():
    real_part = 1.3
    imag_part = 0.3
    complex_value = real_part + 1j * imag_part

    param1 = ComplexParameter("param1_compl", lambda: complex_value, params=None)
    some_value = 3.0 * param1**2 - 1.2j
    true_value = 3.0 * complex_value**2 - 1.2j
    assert true_value == pytest.approx(some_value.numpy(), rel=1e-8)
    assert not param1.get_params()

    # Cartesian complex
    real_part_param = Parameter("real_part_param", real_part)
    imag_part_param = Parameter("imag_part_param", imag_part)
    param2 = ComplexParameter.from_cartesian(
        "param2_compl", real_part_param, imag_part_param
    )
    part1, part2 = param2.get_params()
    part1_val, part2_val = [part1.value().numpy(), part2.value().numpy()]
    if part1_val == pytest.approx(real_part):
        assert part2_val == pytest.approx(imag_part)
    elif part2_val == pytest.approx(real_part):
        assert part1_val == pytest.approx(imag_part)
    else:
        assert False, "one of the if or elif should be the case"

    # Polar complex
    mod_val = 2
    arg_val = np.pi / 4.0
    mod_part_param = Parameter("mod_part_param", mod_val)
    arg_part_param = Parameter("arg_part_param", arg_val)
    param3 = ComplexParameter.from_polar("param3_compl", mod_part_param, arg_part_param)
    part1, part2 = param3.get_cache_deps()
    part1_val, part2_val = [part1.value().numpy(), part2.value().numpy()]
    if part1_val == pytest.approx(mod_val):
        assert part2_val == pytest.approx(arg_val)
    elif part1_val == pytest.approx(arg_val):
        assert part2_val == pytest.approx(mod_val)
    else:
        assert False, "one of the if or elif should be the case"

    param4_name = "param4"
    param4 = ComplexParameter.from_polar(param4_name, 4.0, 2.0, floating=True)
    deps_param4 = param4.get_cache_deps()
    assert len(deps_param4) == 2
    for dep in deps_param4:
        assert dep.floating

    # Test complex conjugate
    param5 = param2.conj

    # Test properties (1e-8 is too precise)
    assert zfit.run(param1.real) == pytest.approx(real_part, rel=1e-6)
    assert zfit.run(param1.imag) == pytest.approx(imag_part, rel=1e-6)
    assert zfit.run(param2.real) == pytest.approx(real_part, rel=1e-6)
    assert zfit.run(param2.imag) == pytest.approx(imag_part, rel=1e-6)
    assert zfit.run(param2.mod) == pytest.approx(np.abs(complex_value))
    assert zfit.run(param2.arg) == pytest.approx(np.angle(complex_value))
    assert zfit.run(param3.mod) == pytest.approx(mod_val, rel=1e-6)
    assert zfit.run(param3.arg) == pytest.approx(arg_val, rel=1e-6)
    assert zfit.run(param3.real) == pytest.approx(mod_val * np.cos(arg_val), rel=1e-6)
    assert zfit.run(param3.imag) == pytest.approx(mod_val * np.sin(arg_val), rel=1e-6)
    assert zfit.run(param5.real) == pytest.approx(real_part)
    assert zfit.run(param5.imag) == pytest.approx(-imag_part)


def test_repr():
    val = 1543
    val2 = 1543**2
    param1 = Parameter("param1", val)
    param2 = zfit.ComposedParameter("comp1", lambda x: x**2, params=param1)
    repr_value = repr(param1)
    repr_value2 = repr(param2)
    assert str(val) in repr_value

    @z.function
    def tf_call():
        repr_value = repr(param1)
        repr_value2 = repr(param2)
        assert str(val) not in repr_value
        assert str(val2) not in repr_value2
        assert "graph-node" in repr_value
        assert "graph-node" in repr_value2

    if zfit.run.get_graph_mode():  # only test if running in graph mode
        tf_call()


def test_composed_param():
    param1 = Parameter("param1", 1.0)
    param2 = Parameter("param2", 2.0)
    param3 = Parameter("param3", 3.0, floating=False)
    param4 = Parameter(
        "param4", 4.0
    )  # noqa Needed to make sure it does not only take all params as deps

    def value_fn(p1, p2, p3):
        return z.math.log(3.0 * p1) * tf.square(p2) - p3

    param_a = ComposedParameter(
        "param_as", value_fn=value_fn, params=(param1, param2, param3)
    )
    param_a2 = ComposedParameter(
        "param_as2",
        value_fn=value_fn,
        params={f"p{i}": p for i, p in enumerate((param1, param2, param3))},
    )
    assert param_a2.params["p1"] == param2
    assert isinstance(param_a.get_cache_deps(only_floating=True), OrderedSet)
    assert param_a.get_cache_deps(only_floating=True) == {param1, param2}
    assert param_a.get_cache_deps(only_floating=False) == {param1, param2, param3}
    a_unchanged = value_fn(param1, param2, param3).numpy()
    assert a_unchanged == param_a.numpy()
    param2.assign(3.5)
    assert param2.numpy()
    a_changed = value_fn(param1, param2, param3).numpy()
    assert a_changed == param_a.numpy()
    assert a_changed != a_unchanged

    # Test param representation
    str(param_a)
    repr(param_a)

    @z.function
    def print_param(p):
        _ = str(p)
        _ = repr(p)

    print_param(param_a)

    # TODO(params): reactivate to check?
    with pytest.raises(LogicalUndefinedOperationError):
        param_a.set_value(value=5.0)
    with pytest.raises(LogicalUndefinedOperationError):
        param_a.assign(value=5.0)
    with pytest.raises(LogicalUndefinedOperationError):
        param_a.randomize()


def test_shape_parameter():
    a = Parameter(name="a", value=1)
    assert a.shape.rank == 0


def test_shape_composed_parameter():
    a = Parameter(name="a", value=1)
    b = Parameter(name="b", value=2)

    def compose():
        return tf.square(a) - b

    c = ComposedParameter(name="c", value_fn=compose, params=[a, b])
    assert c.shape.rank == 0


# TODO: add test
def test_randomize():
    param1 = zfit.Parameter("param1", 1.0, 0, 2)
    for _ in range(100):
        param1.randomize()
        assert 0 < param1 < 2


def test_floating_behavior():
    param1 = zfit.Parameter("param1", 1.0)
    assert param1.floating


def test_param_limits():
    lower, upper = -4.0, 3.0
    param1 = Parameter("param1", 1.0, lower=lower, upper=upper)
    param2 = Parameter("param2", 2.0)

    assert param1.has_limits
    assert not param2.has_limits

    with pytest.raises(ValueError):
        param1.set_value(upper + 0.5)
    param1.assign(upper + 0.5)
    assert upper == param1.value().numpy()
    assert param1.at_limit
    with pytest.raises(ValueError):
        param1.set_value(lower - 1.1)
    param1.assign(lower - 1.1)
    assert lower == param1.value().numpy()
    assert param1.at_limit
    param1.set_value(upper - 0.1)
    assert not param1.at_limit

    param2.lower = lower
    param2.assign(lower - 1.1)
    assert lower == param2.value().numpy()


def test_overloaded_operators():
    param1 = zfit.Parameter("param1", 5)
    param2 = zfit.Parameter("param2", 3)
    param_a = ComposedParameter("param_a", lambda p1: p1 * 4, params=param1)
    param_b = ComposedParameter("param_b", lambda p2: p2, params=param2)
    param_c = param_a * param_b
    assert not isinstance(param_c, zfit.Parameter)
    param_d = ComposedParameter(
        "param_d", lambda pa, pb: pa + pa * pb**2, params=[param_a, param_b]
    )
    param_d_val = param_d.numpy()
    assert param_d_val == (param_a + param_a * param_b**2).numpy()


def test_equal_naming():
    unique_name = "fafdsfds"
    param_unique_name = zfit.Parameter(unique_name, 5.0)
    with pytest.raises(NameAlreadyTakenError):
        param_unique_name2 = zfit.Parameter(unique_name, 3.0)


def test_set_value():
    value1 = 1.0
    value2 = 2.0
    value3 = 3.0
    value4 = 4.0
    param1 = zfit.Parameter(name="param1", value=value1)
    assert param1.numpy() == value1
    with param1.set_value(value2):
        assert param1.numpy() == value2
        param1.set_value(value3)
        assert param1.numpy() == value3
        with param1.set_value(value4):
            assert param1.numpy() == value4
        assert param1.numpy() == value3
    assert param1.numpy() == value1


def test_fixed_param():
    obs = zfit.Space("obs1", (-1, 2))
    sigma = zfit.param.ConstantParameter("const1", 5)
    gauss = zfit.pdf.Gauss(1.0, sigma, obs=obs)
    mu = gauss.params["mu"]
    assert isinstance(mu, zfit.param.ConstantParameter)
    assert isinstance(sigma, zfit.param.ConstantParameter)
    assert not sigma.floating
    assert not sigma.independent
    assert sigma.get_cache_deps() == set()


def test_convert_to_parameters():
    import zfit

    conv_param1 = zfit.param.convert_to_parameters(
        [23, 10.0, 34, 23], prefer_constant=True
    )
    assert len(conv_param1) == 4
    assert not any(p.floating for p in conv_param1)

    conv_param2 = zfit.param.convert_to_parameters(
        [23, 10.0, 34, 12, 23], prefer_constant=False
    )
    assert all(p.floating for p in conv_param2)

    trueval3 = [23, 10.0, 12, 34, 23]
    truelower3 = list(range(5))
    conv_param3 = zfit.param.convert_to_parameters(
        trueval3, lower=truelower3, prefer_constant=False
    )
    np.testing.assert_allclose(zfit.run(conv_param3), trueval3)
    np.testing.assert_allclose(zfit.run([p.lower for p in conv_param3]), truelower3)

    truename4 = ["oe", "myname1", "ue", "eu", "eue"]
    stepsize4 = [23, 1.5, 10.0, 34, 23]
    trueupper4 = [213, 14.0, 1110.0, 314, 213]
    values4 = [23, 12, 121, 34, 23]

    conv_param4dict = zfit.param.convert_to_parameters(
        {
            "value": values4,
            "name": truename4,
            "step_size": stepsize4,
            "upper": trueupper4,
        },
        prefer_constant=False,
    )
    assert [p.name for p in conv_param4dict] == truename4

    np.testing.assert_allclose(zfit.run([p.upper for p in conv_param4dict]), trueupper4)
    np.testing.assert_allclose(
        zfit.run([p.step_size for p in conv_param4dict]), stepsize4
    )

    truename5 = [name + "_five" for name in truename4]
    conv_param4dict = zfit.param.convert_to_parameters(
        values4,
        name=truename5,
        upper=trueupper4,
        prefer_constant=False,
        step_size=stepsize4,
    )
    assert [p.name for p in conv_param4dict] == truename5

    np.testing.assert_allclose(zfit.run([p.upper for p in conv_param4dict]), trueupper4)
    np.testing.assert_allclose(
        zfit.run([p.step_size for p in conv_param4dict]), stepsize4
    )


def test_convert_to_parameters_equivalence_to_single_multi():
    import zfit

    conv_param1 = zfit.param.convert_to_parameters([23, 10.0, 34, 23])[1]
    assert pytest.approx(zfit.run(conv_param1.value())) == 10
    assert not conv_param1.floating

    conv_param2 = zfit.param.convert_to_parameters(
        [23, 10.0, 34, 12, 23], prefer_constant=False
    )[-2]
    assert pytest.approx(zfit.run(conv_param2.value())) == 12.0
    assert conv_param2.floating
    assert not conv_param2.has_limits

    conv_param3 = zfit.param.convert_to_parameters(
        [23, 10.0, 12, 34, 23], lower=list(range(5)), prefer_constant=False
    )[2]
    assert pytest.approx(zfit.run(conv_param3.lower)) == 2
    assert conv_param3.has_limits

    truename4 = ["oe", "myname1", "ue", "eu", "eue"]
    stepsize4 = [23, 1.5, 10.0, 34, 23]
    conv_param4 = zfit.param.convert_to_parameters(
        [23, 12, 121, 34, 23],
        name=truename4,
        upper=[213, 14.0, 1110.0, 314, 213],
        prefer_constant=False,
        step_size=stepsize4,
    )[1]
    assert conv_param4.floating
    assert conv_param4.name == "myname1"
    assert conv_param4.has_limits
    assert conv_param4.floating
    assert pytest.approx(zfit.run(conv_param4.step_size)) == 1.5


def test_convert_to_parameters_equivalence_to_single():
    import zfit

    conv_param1 = zfit.param.convert_to_parameters(10.0)[0]
    assert pytest.approx(zfit.run(conv_param1.value())) == 10
    assert not conv_param1.floating

    conv_param2 = zfit.param.convert_to_parameters(12.0, prefer_constant=False)[0]
    assert pytest.approx(zfit.run(conv_param2.value())) == 12.0
    assert conv_param2.floating
    assert not conv_param2.has_limits

    conv_param3 = zfit.param.convert_to_parameters(
        12.0, lower=5.0, prefer_constant=False
    )[0]
    assert pytest.approx(zfit.run(conv_param3.lower)) == 5.0
    assert conv_param3.has_limits

    truename4 = "myname1"
    stepsize4 = 1.5
    conv_param4 = zfit.param.convert_to_parameters(
        12.0, name=truename4, upper=14.0, prefer_constant=False, step_size=stepsize4
    )[0]
    assert conv_param4.floating
    assert conv_param4.name == truename4
    assert conv_param4.has_limits
    assert conv_param4.floating
    assert pytest.approx(zfit.run(conv_param4.step_size)) == stepsize4


def test_convert_to_parameter():
    import zfit

    conv_param1 = zfit.param.convert_to_parameter(10.0)
    assert pytest.approx(zfit.run(conv_param1.value())) == 10
    assert not conv_param1.floating

    conv_param2 = zfit.param.convert_to_parameter(12.0, prefer_constant=False)
    assert pytest.approx(zfit.run(conv_param2.value())) == 12.0
    assert conv_param2.floating
    assert not conv_param2.has_limits

    conv_param3 = zfit.param.convert_to_parameter(
        12.0, lower=5.0, prefer_constant=False
    )
    assert pytest.approx(zfit.run(conv_param3.lower)) == 5.0
    assert conv_param3.has_limits

    with pytest.raises(ValueError):
        _ = zfit.param.convert_to_parameter(5.0, lower=15.0, prefer_constant=False)

    with pytest.raises(ValueError):
        _ = zfit.param.convert_to_parameter(5.0, upper=1.0, prefer_constant=False)

    truename4 = "myname1"
    stepsize4 = 1.5
    conv_param4 = zfit.param.convert_to_parameter(
        12.0, name=truename4, upper=14.0, prefer_constant=False, step_size=stepsize4
    )
    assert conv_param4.floating
    assert conv_param4.name == truename4
    assert conv_param4.has_limits
    assert conv_param4.floating
    assert pytest.approx(zfit.run(conv_param4.step_size)) == stepsize4


def test_set_values():
    import zfit

    init_values = [1, 2, 3]
    second_values = [5, 6, 7]
    params = [zfit.Parameter(f"param_{i}", val) for i, val in enumerate(init_values)]

    with zfit.param.set_values(params, second_values):
        for param, val in zip(params, second_values):
            assert param.value().numpy() == val

    for param, val in zip(params, init_values):
        assert param.value().numpy() == val

    zfit.param.set_values(params, second_values)
    for param, val in zip(params, second_values):
        assert param.value().numpy() == val

    zfit.param.set_values(params, init_values)
    for param, val in zip(params, init_values):
        assert param.value().numpy() == val


@pytest.mark.parametrize("addmore", [True, False])
def test_set_values_dict(addmore):
    import zfit

    init_values = [1, 2, 3]
    second_values = [5, 6, 7]
    params = [zfit.Parameter(f"param_{i}", val) for i, val in enumerate(init_values)]

    setvalueparam = {p: v for p, v in zip(params, second_values)}
    setvalue_paramname = {p.name: v for p, v in zip(params, second_values)}
    setvaluemixed = setvalueparam.copy()
    setvaluemixed["param_1"] = params[1]

    if addmore:
        param4 = zfit.Parameter("param4", 4)
        setvalue_paramname[param4.name] = 4
        setvalueparam[param4] = 4
        setvaluemixed[param4] = 4

    zfit.param.set_values(params, setvalueparam)
    for param, val in zip(params, second_values):
        assert pytest.approx(param.value().numpy()) == val

    zfit.param.set_values(params, init_values)

    zfit.param.set_values(params, setvalue_paramname)
    for param, val in zip(params, second_values):
        assert pytest.approx(param.value().numpy()) == val

    zfit.param.set_values(params, init_values)
    zfit.param.set_values(params, setvaluemixed)
    for param, val in zip(params, second_values):
        assert pytest.approx(param.value().numpy()) == val

    zfit.param.set_values(params, init_values)

    too_small_values = setvalueparam.copy()
    too_small_values.pop(params[0])
    with pytest.raises(ValueError):
        zfit.param.set_values(params, too_small_values)

    zfit.param.set_values(params, init_values)
    # try with allow_partial
    zfit.param.set_values(params, too_small_values, allow_partial=True)
    assert params[0].value().numpy() == init_values[0]
    assert params[1].value().numpy() == second_values[1]
    assert params[2].value().numpy() == second_values[2]

    zfit.param.set_values(params, init_values)


def test_deletion():
    import gc

    def func():
        a = zfit.Parameter("param", 42)
        return True

    assert func()

    gc.collect()
    assert func()  # this must not raise an error


def test_to_numpy():
    import zfit

    param = zfit.Parameter("param", 42)
    assert zfit.run(param) == 42

    p1 = zfit.param.ConstantParameter("aoeu1", 15)
    assert zfit.run(p1) == 15

    p2 = zfit.param.ComposedParameter(
        "aoeu2", lambda params: 2 * params["p1"], {"p1": p1}
    )
    assert zfit.run(p2) == 30



#  Copyright (c) 2022 zfit
import numpy as np
import pytest
import tensorflow as tf

from zfit import Parameter, z
from zfit.models.functions import SimpleFuncV1
from zfit.models.special import SimplePDF
from zfit.util.exception import BreakingAPIChangeError, ModelIncompatibleError

rnd_test_values = np.array([1.0, 0.01, -14.2, 0.0, 1.5, 152, -0.1, 12])

obs1 = "obs1"


def test_not_allowed():
    param1 = Parameter("param1", 1.0)
    param2 = Parameter("param2", 2.0)
    param3 = Parameter("param3", 3.0, floating=False)

    def func1_pure(self, x):
        return param1 * x

    def func2_pure(self, x):
        return param2 * x + param3

    func1 = SimpleFuncV1(func=func1_pure, obs=obs1, p1=param1)

    pdf1 = SimplePDF(func=lambda self, x: x * param1, obs=obs1)
    pdf2 = SimplePDF(func=lambda self, x: x * param2, obs=obs1)

    with pytest.raises(BreakingAPIChangeError):
        pdf1 + pdf2

    ext_pdf1 = pdf1.create_extended(param1)
    with pytest.raises(BreakingAPIChangeError):
        ext_pdf1 + pdf2
    with pytest.raises(BreakingAPIChangeError):
        param2 * pdf1
    with pytest.raises(NotImplementedError):
        param1 + func1
    with pytest.raises(NotImplementedError):
        func1 + param1
    with pytest.raises(ModelIncompatibleError):
        func1 * pdf2
    with pytest.raises(ModelIncompatibleError):
        pdf1 * func1


def test_param_func():
    param1 = Parameter("param1", 1.0)
    param2 = Parameter("param2", 2.0)
    param3 = Parameter("param3", 3.0, floating=False)
    param4 = Parameter("param4", 4.0)
    a = z.math.log(3.0 * param1) * tf.square(param2) - param3
    func = SimpleFuncV1(func=lambda self, x: a * x.value()[:, 0], obs=obs1)

    new_func = param4 * func

    new_func_equivalent = func * param4

    result1 = new_func.func(x=rnd_test_values).numpy()
    result1_equivalent = new_func_equivalent.func(x=rnd_test_values).numpy()
    result2 = func.func(x=rnd_test_values) * param4
    np.testing.assert_array_equal(result1, result2)
    np.testing.assert_array_equal(result1_equivalent, result2)


def test_func_func():
    param1 = Parameter("param1", 1.0)
    param2 = Parameter("param2", 2.0)
    param3 = Parameter("param3", 3.0, floating=False)

    def func1_pure(self, x):
        x = z.unstack_x(x)
        return param1 * x

    def func2_pure(self, x):
        x = z.unstack_x(x)
        return param2 * x + param3

    func1 = SimpleFuncV1(func=func1_pure, obs=obs1, p1=param1)
    func2 = SimpleFuncV1(func=func2_pure, obs=obs1, p2=param2, p3=param3)

    added_func = func1 + func2
    prod_func = func1 * func2

    added_values = added_func.func(rnd_test_values)
    true_added_values = func1_pure(None, rnd_test_values) + func2_pure(
        None, rnd_test_values
    )
    prod_values = prod_func.func(rnd_test_values)
    true_prod_values = func1_pure(None, rnd_test_values) * func2_pure(
        None, rnd_test_values
    )

    added_values = added_values.numpy()
    true_added_values = true_added_values.numpy()
    prod_values = prod_values.numpy()
    true_prod_values = true_prod_values.numpy()
    np.testing.assert_allclose(true_added_values, added_values)
    np.testing.assert_allclose(true_prod_values, prod_values)



#  Copyright (c) 2023 zfit
"""Recurrent polynomials."""
from __future__ import annotations

from typing import TYPE_CHECKING, Optional

import pydantic

from typing import Literal

from ..core.serialmixin import SerializableMixin
from ..serialization import SpaceRepr, Serializer
from ..serialization.pdfrepr import BasePDFRepr
from ..util.ztyping import ExtendedInputType, NormInputType

if TYPE_CHECKING:
    import zfit

from typing import Mapping
import abc

import tensorflow as tf

import zfit.z.numpy as znp
from zfit import z

from ..core.basepdf import BasePDF
from ..core.space import Space
from ..settings import ztypes
from ..util import ztyping
from ..util.container import convert_to_container
from ..util.exception import SpecificFunctionNotImplemented


def rescale_minus_plus_one(x: tf.Tensor, limits: zfit.Space) -> tf.Tensor:
    """Rescale and shift *x* as *limits* were rescaled and shifted to be in (-1, 1). Useful for orthogonal polynomials.

    Args:
        x: Array like data
        limits: 1-D limits

    Returns:
        The rescaled tensor.
    """
    lim_low, lim_high = limits.limit1d
    x = (2 * x - lim_low - lim_high) / (lim_high - lim_low)
    return x


class RecursivePolynomial(BasePDF):
    """1D polynomial generated via three-term recurrence."""

    def __init__(
        self,
        obs,
        coeffs: list,
        apply_scaling: bool = True,
        coeff0: tf.Tensor | None = None,
        *,
        extended: ExtendedInputType = None,
        norm: NormInputType = None,
        name: str = "Polynomial",
    ):  # noqa
        """Base class to create 1 dimensional recursive polynomials that can be rescaled. Overwrite _poly_func.

        Args:
            obs: |@doc:pdf.init.obs| Observables of the
               model. This will be used as the default space of the PDF and,
               if not given explicitly, as the normalization range.

               The default space is used for example in the sample method: if no
               sampling limits are given, the default space is used.

               The observables are not equal to the domain as it does not restrict or
               truncate the model outside this range. |@docend:pdf.init.obs|
            coeffs: |@doc:pdf.polynomial.init.coeffs| Coefficients of the sum of the polynomial.
               The coefficients of the polynomial, starting with the first order
               term. To set the constant term, use ``coeff0``. |@docend:pdf.polynomial.init.coeffs|
            apply_scaling: |@doc:pdf.polynomial.init.apply_scaling| Rescale the data so that the actual limits represent (-1, 1).
               This is usually wanted as the polynomial is defined in this range.
               Default is ``True``. |@docend:pdf.polynomial.init.apply_scaling|

                .. math::
                   x_{n+1} = recurrence(x_{n}, x_{n-1}, n)

            extended: |@doc:pdf.init.extended| The overall yield of the PDF.
               If this is parameter-like, it will be used as the yield,
               the expected number of events, and the PDF will be extended.
               An extended PDF has additional functionality, such as the
               ``ext_*`` methods and the ``counts`` (for binned PDFs). |@docend:pdf.init.extended|
            norm: |@doc:pdf.init.norm| Normalization of the PDF.
               By default, this is the same as the default space of the PDF. |@docend:pdf.init.norm|
        """
        # 0th coefficient set to 1 by default
        coeff0 = (
            z.constant(1.0) if coeff0 is None else tf.cast(coeff0, dtype=ztypes.float)
        )
        coeffs = convert_to_container(coeffs).copy()
        coeffs.insert(0, coeff0)
        params = {f"c_{i}": coeff for i, coeff in enumerate(coeffs)}
        self._degree = len(coeffs) - 1  # 1 coeff -> 0th degree
        self._apply_scale = apply_scaling
        if apply_scaling and not (isinstance(obs, Space) and obs.n_limits == 1):
            raise ValueError(
                "obs need to be a Space with exactly one limit if rescaling is requested."
            )
        super().__init__(
            obs=obs, name=name, params=params, extended=extended, norm=norm
        )

    def _polynomials_rescale(self, x):
        if self._apply_scale:
            x = rescale_minus_plus_one(x, limits=self.space)
        return x

    @property
    def apply_scaling(self):
        return self._apply_scale

    @property
    def degree(self):
        """Int: degree of the polynomial, starting from 0."""
        return self._degree

    def _unnormalized_pdf(self, x):
        x = x.unstack_x()
        x = self._polynomials_rescale(x)
        return self._poly_func(x=x)

    @abc.abstractmethod
    def _poly_func(self, x):
        raise SpecificFunctionNotImplemented


class BaseRecursivePolynomialRepr(BasePDFRepr):
    x: SpaceRepr
    params: Mapping[str, Serializer.types.ParamTypeDiscriminated] = pydantic.Field(
        alias="coeffs"
    )
    apply_scaling: Optional[bool]

    @pydantic.root_validator(pre=True)
    def convert_params(cls, values):  # does not propagate `params` into the fields
        if cls.orm_mode(values):
            values = dict(values)
            values["x"] = values.pop("space")
        return values

    def _to_orm(self, init):
        init["coeff0"], *init["coeffs"] = init.pop("params").values()
        return super()._to_orm(init)


def create_poly(x, polys, coeffs, recurrence):
    degree = len(coeffs) - 1
    polys = do_recurrence(x, polys=polys, degree=degree, recurrence=recurrence)
    sum_polys = znp.sum([coeff * poly for coeff, poly in zip(coeffs, polys)], axis=0)
    return sum_polys


def do_recurrence(x, polys, degree, recurrence):
    polys = [polys[0](x), polys[1](x)]
    for i_deg in range(1, degree):  # recurrence returns for the n+1th degree
        polys.append(recurrence(polys[-1], polys[-2], i_deg, x))
    return polys


legendre_polys = [lambda x: tf.ones_like(x), lambda x: x]


@z.function(wraps="zfit_tensor", stateless_args=False)
def legendre_recurrence(p1, p2, n, x):
    """Recurrence relation for Legendre polynomials.

    .. math::
         (n+1) P_{n+1}(x) = (2n + 1) x P_{n}(x) - n P_{n-1}(x)
    """
    return ((2 * n + 1) * znp.multiply(x, p1) - n * p2) / (n + 1)


def legendre_shape(x, coeffs):
    return create_poly(
        x=x, polys=legendre_polys, coeffs=coeffs, recurrence=legendre_recurrence
    )


def legendre_integral(
    limits: ztyping.SpaceType,
    norm: ztyping.SpaceType,
    params: list[zfit.Parameter],
    model: RecursivePolynomial,
):
    """Recursive integral of Legendre polynomials."""
    lower, upper = limits.limit1d
    lower_rescaled = model._polynomials_rescale(lower)
    upper_rescaled = model._polynomials_rescale(upper)
    # if np.allclose((lower_rescaled, upper_rescaled), (-1, 1)):
    #     return z.constant(2.)  #

    lower = z.convert_to_tensor(lower_rescaled)
    upper = z.convert_to_tensor(upper_rescaled)

    integral_0 = model.params["c_0"] * (upper - lower)  # if polynomial 0 is 1
    if model.degree == 0:
        integral = integral_0
    else:

        def indefinite_integral(limits):
            max_degree = (
                model.degree + 1
            )  # needed +1 for integral, max poly in term for n is n+1
            polys = do_recurrence(
                x=limits,
                polys=legendre_polys,
                degree=max_degree,
                recurrence=legendre_recurrence,
            )
            one_limit_integrals = []
            for degree in range(1, max_degree):
                coeff = model.params[f"c_{degree}"]
                one_limit_integrals.append(
                    coeff
                    * (polys[degree + 1] - polys[degree - 1])
                    / (2.0 * (z.convert_to_tensor(degree)) + 1)
                )
            return z.reduce_sum(one_limit_integrals, axis=0)

        integral = indefinite_integral(upper) - indefinite_integral(lower) + integral_0
        integral = znp.reshape(integral, newshape=())
    integral *= 0.5 * model.space.area()  # rescale back to whole width

    return integral


class Legendre(RecursivePolynomial, SerializableMixin):
    def __init__(
        self,
        obs: ztyping.ObsTypeInput,
        coeffs: list[ztyping.ParamTypeInput],
        apply_scaling: bool = True,
        coeff0: ztyping.ParamTypeInput | None = None,
        *,
        extended: ExtendedInputType = None,
        norm: NormInputType = None,
        name: str = "Legendre",
    ):  # noqa
        """Linear combination of Legendre polynomials of order len(coeffs), the coeffs are overall scaling factors.

        The 0th coefficient is set to 1 by default but can be explicitly set with *coeff0*. Since the PDF normalization
        removes a degree of freedom, the 0th coefficient is redundant and leads to an arbitrary overall scaling of all
        parameters.

        Notice that this is already a sum of polynomials and the coeffs are simply scaling the individual orders of the
        polynomials.

        The recursive definition of **a single order** of the polynomial is

        .. math::
            (n+1) P_{n+1}(x) = (2n + 1) x P_{n}(x) - n P_{n-1}(x)

            with
            P_0 = 1
            P_1 = x


        Args:
            obs: |@doc:pdf.init.obs| Observables of the
               model. This will be used as the default space of the PDF and,
               if not given explicitly, as the normalization range.

               The default space is used for example in the sample method: if no
               sampling limits are given, the default space is used.

               The observables are not equal to the domain as it does not restrict or
               truncate the model outside this range. |@docend:pdf.init.obs|
            coeffs: A list of the coefficients for the polynomials of order 1+ in the sum.
            apply_scaling: Rescale the data so that the actual limits represent (-1, 1).
            coeff0: The scaling factor of the 0th order polynomial. If not given, it is set to 1.
            extended: |@doc:pdf.init.extended| The overall yield of the PDF.
               If this is parameter-like, it will be used as the yield,
               the expected number of events, and the PDF will be extended.
               An extended PDF has additional functionality, such as the
               ``ext_*`` methods and the ``counts`` (for binned PDFs). |@docend:pdf.init.extended|
            norm: |@doc:pdf.init.norm| Normalization of the PDF.
               By default, this is the same as the default space of the PDF. |@docend:pdf.init.norm|
            name: |@doc:pdf.init.name| Human-readable name
               or label of
               the PDF for better identification.
               Has no programmatical functional purpose as identification. |@docend:pdf.init.name|
        """
        super().__init__(
            obs=obs,
            name=name,
            coeffs=coeffs,
            apply_scaling=apply_scaling,
            coeff0=coeff0,
            extended=extended,
            norm=norm,
        )

    def _poly_func(self, x):
        coeffs = convert_coeffs_dict_to_list(self.params)
        return legendre_shape(x=x, coeffs=coeffs)


class LegendreRepr(BaseRecursivePolynomialRepr):
    _implementation = Legendre
    hs3_type: Literal["Legendre"] = pydantic.Field("Legendre", alias="type")


legendre_limits = Space(axes=0, limits=(Space.ANY_LOWER, Space.ANY_UPPER))
Legendre.register_analytic_integral(func=legendre_integral, limits=legendre_limits)

chebyshev_polys = [lambda x: tf.ones_like(x), lambda x: x]


@z.function(wraps="zfit_tensor", stateless_args=False)
def chebyshev_recurrence(p1, p2, _, x):
    """Recurrence relation for Chebyshev polynomials.

    T_{n+1}(x) = 2 x T_{n}(x) - T_{n-1}(x)
    """
    return 2 * znp.multiply(x, p1) - p2


def chebyshev_shape(x, coeffs):
    return create_poly(
        x=x, polys=chebyshev_polys, coeffs=coeffs, recurrence=chebyshev_recurrence
    )


class Chebyshev(RecursivePolynomial, SerializableMixin):
    def __init__(
        self,
        obs,
        coeffs: list,
        apply_scaling: bool = True,
        coeff0: ztyping.ParamTypeInput | None = None,
        *,
        extended: ExtendedInputType = None,
        norm: NormInputType = None,
        name: str = "Chebyshev",
    ):  # noqa
        """Linear combination of Chebyshev (first kind) polynomials of order len(coeffs), coeffs are scaling factors.

        The 0th coefficient is set to 1 by default but can be explicitly set with *coeff0*. Since the PDF normalization
        removes a degree of freedom, the 0th coefficient is redundant and leads to an arbitrary overall scaling of all
        parameters.

        Notice that this is already a sum of polynomials and the coeffs are simply scaling the individual orders of the
        polynomials.

        The recursive definition of **a single order** of the polynomial is

        .. math::
            T_{n+1}(x) = 2 x T_{n}(x) - T_{n-1}(x)

            with
            T_{0} = 1
            T_{1} = x

        Notice that :math:`T_1` is x as opposed to 2x in Chebyshev polynomials of the second kind.

        Args:
            obs: |@doc:pdf.init.obs| Observables of the
               model. This will be used as the default space of the PDF and,
               if not given explicitly, as the normalization range.

               The default space is used for example in the sample method: if no
               sampling limits are given, the default space is used.

               The observables are not equal to the domain as it does not restrict or
               truncate the model outside this range. |@docend:pdf.init.obs|
            coeffs: A list of the coefficients for the polynomials of order 1+ in the sum.
            apply_scaling: Rescale the data so that the actual limits represent (-1, 1).
            coeff0: The scaling factor of the 0th order polynomial. If not given, it is set to 1.
            extended: |@doc:pdf.init.extended| The overall yield of the PDF.
               If this is parameter-like, it will be used as the yield,
               the expected number of events, and the PDF will be extended.
               An extended PDF has additional functionality, such as the
               ``ext_*`` methods and the ``counts`` (for binned PDFs). |@docend:pdf.init.extended|
            norm: |@doc:pdf.init.norm| Normalization of the PDF.
               By default, this is the same as the default space of the PDF. |@docend:pdf.init.norm|
            name: |@doc:pdf.init.name| Human-readable name
               or label of
               the PDF for better identification.
               Has no programmatical functional purpose as identification. |@docend:pdf.init.name|
        """
        super().__init__(
            obs=obs,
            name=name,
            coeffs=coeffs,
            coeff0=coeff0,
            apply_scaling=apply_scaling,
            extended=extended,
            norm=norm,
        )

    def _poly_func(self, x):
        coeffs = convert_coeffs_dict_to_list(self.params)
        return chebyshev_shape(x=x, coeffs=coeffs)


class ChebyshevRepr(BaseRecursivePolynomialRepr):
    _implementation = Chebyshev
    hs3_type: Literal["Chebyshev"] = pydantic.Field("Chebyshev", alias="type")


def func_integral_chebyshev1(limits, norm, params, model):
    lower, upper = limits.rect_limits
    lower_rescaled = model._polynomials_rescale(lower)
    upper_rescaled = model._polynomials_rescale(upper)

    lower = z.convert_to_tensor(lower_rescaled)
    upper = z.convert_to_tensor(upper_rescaled)

    integral = model.params["c_0"] * (
        upper - lower
    )  # if polynomial 0 is defined as T_0 = 1
    if model.degree >= 1:
        integral += (
            model.params["c_1"] * 0.5 * (upper**2 - lower**2)
        )  # if polynomial 0 is defined as T_0 = 1
    if model.degree >= 2:

        def indefinite_integral(limits):
            max_degree = model.degree + 1
            polys = do_recurrence(
                x=limits,
                polys=chebyshev_polys,
                degree=max_degree,
                recurrence=chebyshev_recurrence,
            )
            one_limit_integrals = []
            for degree in range(2, max_degree):
                coeff = model.params[f"c_{degree}"]
                n_float = z.convert_to_tensor(degree)
                integral = n_float * polys[degree + 1] / (
                    z.square(n_float) - 1
                ) - limits * polys[degree] / (n_float - 1)
                one_limit_integrals.append(coeff * integral)
            return z.reduce_sum(one_limit_integrals, axis=0)

        integral += indefinite_integral(upper) - indefinite_integral(lower)
        integral = znp.reshape(integral, newshape=())
    integral *= 0.5 * model.space.area()  # rescale back to whole width
    integral = tf.gather(integral, indices=0, axis=-1)
    return integral


chebyshev1_limits_integral = Space(axes=0, limits=(Space.ANY_LOWER, Space.ANY_UPPER))
Chebyshev.register_analytic_integral(
    func=func_integral_chebyshev1, limits=chebyshev1_limits_integral
)

chebyshev2_polys = [lambda x: tf.ones_like(x), lambda x: x * 2]


def chebyshev2_shape(x, coeffs):
    return create_poly(
        x=x, polys=chebyshev2_polys, coeffs=coeffs, recurrence=chebyshev_recurrence
    )


class Chebyshev2(RecursivePolynomial, SerializableMixin):
    def __init__(
        self,
        obs,
        coeffs: list,
        apply_scaling: bool = True,
        coeff0: ztyping.ParamTypeInput | None = None,
        *,
        extended: ExtendedInputType = None,
        norm: NormInputType = None,
        name: str = "Chebyshev2",
    ):  # noqa
        """Linear combination of Chebyshev (second kind) polynomials of order len(coeffs), coeffs are scaling factors.

        The 0th coefficient is set to 1 by default but can be explicitly set with *coeff0*. Since the PDF normalization
        removes a degree of freedom, the 0th coefficient is redundant and leads to an arbitrary overall scaling of all
        parameters.

        Notice that this is already a sum of polynomials and the coeffs are simply scaling the individual orders of the
        polynomials.

        The recursive definition of **a single order** of the polynomial is

        .. math::
            T_{n+1}(x) = 2 x T_{n}(x) - T_{n-1}(x)

            with
            T_{0} = 1
            T_{1} = 2x

        Notice that :math:`T_1` is 2x as opposed to x in Chebyshev polynomials of the first kind.



        Args:
            obs: The default space the PDF is defined in.
            coeffs: A list of the coefficients for the polynomials of order 1+ in the sum.
            apply_scaling: Rescale the data so that the actual limits represent (-1, 1).
            coeff0: The scaling factor of the 0th order polynomial. If not given, it is set to 1.
            extended: |@doc:pdf.init.extended| The overall yield of the PDF.
               If this is parameter-like, it will be used as the yield,
               the expected number of events, and the PDF will be extended.
               An extended PDF has additional functionality, such as the
               ``ext_*`` methods and the ``counts`` (for binned PDFs). |@docend:pdf.init.extended|
            norm: |@doc:pdf.init.norm| Normalization of the PDF.
               By default, this is the same as the default space of the PDF. |@docend:pdf.init.norm|
            name: Name of the polynomial
        """
        super().__init__(
            obs=obs,
            name=name,
            coeffs=coeffs,
            coeff0=coeff0,
            apply_scaling=apply_scaling,
            extended=extended,
            norm=norm,
        )

    def _poly_func(self, x):
        coeffs = convert_coeffs_dict_to_list(self.params)
        return chebyshev2_shape(x=x, coeffs=coeffs)


class Chebyshev2Repr(BaseRecursivePolynomialRepr):
    _implementation = Chebyshev2
    hs3_type: Literal["Chebyshev2"] = pydantic.Field("Chebyshev2", alias="type")


def func_integral_chebyshev2(limits, norm, params, model):
    lower, upper = limits.limit1d
    lower_rescaled = model._polynomials_rescale(lower)
    upper_rescaled = model._polynomials_rescale(upper)

    lower = z.convert_to_tensor(lower_rescaled)
    upper = z.convert_to_tensor(upper_rescaled)

    # the integral of cheby2_ni is a cheby1_ni+1/(n+1). We add the (n+1) to the coeffs. The cheby1 shape makes
    # the sum for us.
    coeffs_cheby1 = {"c_0": z.constant(0.0, dtype=model.dtype)}

    for name, coeff in params.items():
        n_plus1 = int(name.split("_", 1)[-1]) + 1
        coeffs_cheby1[f"c_{n_plus1}"] = coeff / z.convert_to_tensor(
            n_plus1, dtype=model.dtype
        )
    coeffs_cheby1 = convert_coeffs_dict_to_list(coeffs_cheby1)

    def indefinite_integral(limits):
        return chebyshev_shape(x=limits, coeffs=coeffs_cheby1)

    integral = indefinite_integral(upper) - indefinite_integral(lower)
    integral = znp.reshape(integral, newshape=())
    integral *= 0.5 * model.space.area()  # rescale back to whole width

    return integral


chebyshev2_limits_integral = Space(axes=0, limits=(Space.ANY_LOWER, Space.ANY_UPPER))
Chebyshev2.register_analytic_integral(
    func=func_integral_chebyshev2, limits=chebyshev2_limits_integral
)


def generalized_laguerre_polys_factory(alpha=0.0):
    return [lambda x: tf.ones_like(x), lambda x: 1 + alpha - x]


laguerre_polys = generalized_laguerre_polys_factory(alpha=0.0)


def generalized_laguerre_recurrence_factory(alpha=0.0):
    @z.function(wraps="zfit_tensor", stateless_args=False)
    def generalized_laguerre_recurrence(p1, p2, n, x):
        """Recurrence relation for Laguerre polynomials.

        :math:`(n+1) L_{n+1}(x) = (2n + 1 + \alpha - x) L_{n}(x) - (n + \alpha) L_{n-1}(x)`
        """
        return (znp.multiply(2 * n + 1 + alpha - x, p1) - (n + alpha) * p2) / (n + 1)

    return generalized_laguerre_recurrence


laguerre_recurrence = generalized_laguerre_recurrence_factory(alpha=0.0)


def generalized_laguerre_shape_factory(alpha=0.0):
    recurrence = generalized_laguerre_recurrence_factory(alpha=alpha)
    polys = generalized_laguerre_polys_factory(alpha=alpha)

    def general_laguerre_shape(x, coeffs):
        return create_poly(x=x, polys=polys, coeffs=coeffs, recurrence=recurrence)

    return general_laguerre_shape


laguerre_shape = generalized_laguerre_shape_factory(alpha=0.0)
laguerre_shape_alpha_minusone = generalized_laguerre_shape_factory(
    alpha=-1.0
)  # for integral


class Laguerre(RecursivePolynomial, SerializableMixin):
    def __init__(
        self,
        obs,
        coeffs: list,
        apply_scaling: bool = True,
        coeff0: ztyping.ParamTypeInput | None = None,
        *,
        extended: ExtendedInputType = None,
        norm: NormInputType = None,
        name: str = "Laguerre",
    ):  # noqa
        """Linear combination of Laguerre polynomials of order len(coeffs), the coeffs are overall scaling factors.

        The 0th coefficient is set to 1 by default but can be explicitly set with *coeff0*. Since the PDF normalization
        removes a degree of freedom, the 0th coefficient is redundant and leads to an arbitrary overall scaling of all
        parameters.

        Notice that this is already a sum of polynomials and the coeffs are simply scaling the individual orders of the
        polynomials.

        The recursive definition of **a single order** of the polynomial is

        .. math::

            (n+1) L_{n+1}(x) = (2n + 1 + \alpha - x) L_{n}(x) - (n + \alpha) L_{n-1}(x)

        with
        P_0 = 1
        P_1 = 1 - x


        Args:
            obs: The default space the PDF is defined in.
            coeffs: A list of the coefficients for the polynomials of order 1+ in the sum.
            apply_scaling: Rescale the data so that the actual limits represent (-1, 1).
            coeff0: The scaling factor of the 0th order polynomial. If not given, it is set to 1.
            extended: |@doc:pdf.init.extended| The overall yield of the PDF.
               If this is parameter-like, it will be used as the yield,
               the expected number of events, and the PDF will be extended.
               An extended PDF has additional functionality, such as the
               ``ext_*`` methods and the ``counts`` (for binned PDFs). |@docend:pdf.init.extended|
            norm: |@doc:pdf.init.norm| Normalization of the PDF.
               By default, this is the same as the default space of the PDF. |@docend:pdf.init.norm|
            name: Name of the polynomial
        """
        super().__init__(
            obs=obs,
            name=name,
            coeffs=coeffs,
            coeff0=coeff0,
            apply_scaling=apply_scaling,
            extended=extended,
            norm=norm,
        )

    def _poly_func(self, x):
        coeffs = convert_coeffs_dict_to_list(self.params)
        return laguerre_shape(x=x, coeffs=coeffs)


class LaguerreRepr(BaseRecursivePolynomialRepr):
    _implementation = Laguerre
    hs3_type: Literal["Laguerre"] = pydantic.Field("Laguerre", alias="type")


def func_integral_laguerre(limits, norm, params: dict, model):
    """The integral of the simple laguerre polynomials.

    Defined as :math:`\\int L_{n} = (-1) L_{n+1}^{(-1)}` with :math:`L^{(\alpha)}` the generalized Laguerre polynom.

    Args:
        limits:
        norm:
        params:
        model:

    Returns:
    """
    lower, upper = limits.limit1d
    lower_rescaled = model._polynomials_rescale(lower)
    upper_rescaled = model._polynomials_rescale(upper)

    lower = z.convert_to_tensor(lower_rescaled)
    upper = z.convert_to_tensor(upper_rescaled)

    # The laguerre shape makes the sum for us. setting the 0th coeff to 0, since no -1 term exists.
    coeffs_laguerre_nup = {
        f'c_{int(n.split("_", 1)[-1]) + 1}': c
        for i, (n, c) in enumerate(params.items())
    }  # increase n -> n+1 of naming
    coeffs_laguerre_nup["c_0"] = tf.constant(0.0, dtype=model.dtype)
    coeffs_laguerre_nup = convert_coeffs_dict_to_list(coeffs_laguerre_nup)

    def indefinite_integral(limits):
        return -1 * laguerre_shape_alpha_minusone(x=limits, coeffs=coeffs_laguerre_nup)

    integral = indefinite_integral(upper) - indefinite_integral(lower)
    integral = znp.reshape(integral, newshape=())
    integral *= 0.5 * model.space.area()  # rescale back to whole width
    return integral


laguerre_limits_integral = Space(axes=0, limits=(Space.ANY_LOWER, Space.ANY_UPPER))
Laguerre.register_analytic_integral(
    func=func_integral_laguerre, limits=laguerre_limits_integral
)

hermite_polys = [lambda x: tf.ones_like(x), lambda x: 2 * x]


@z.function(wraps="zfit_tensor", stateless_args=False)
def hermite_recurrence(p1, p2, n, x):
    """Recurrence relation for Hermite polynomials (physics).

    :math:`H_{n+1}(x) = 2x H_{n}(x) - 2n H_{n-1}(x)`
    """
    return 2 * (znp.multiply(x, p1) - n * p2)


def hermite_shape(x, coeffs):
    return create_poly(
        x=x, polys=hermite_polys, coeffs=coeffs, recurrence=hermite_recurrence
    )


class Hermite(RecursivePolynomial, SerializableMixin):
    def __init__(
        self,
        obs,
        coeffs: list,
        apply_scaling: bool = True,
        coeff0: ztyping.ParamTypeInput | None = None,
        *,
        extended: ExtendedInputType = None,
        norm: NormInputType = None,
        name: str = "Hermite",
    ):  # noqa
        """Linear combination of Hermite polynomials (for physics) of order len(coeffs), with coeffs as scaling factors.

        The 0th coefficient is set to 1 by default but can be explicitly set with *coeff0*. Since the PDF normalization
        removes a degree of freedom, the 0th coefficient is redundant and leads to an arbitrary overall scaling of all
        parameters.

        Notice that this is already a sum of polynomials and the coeffs are simply scaling the individual orders of the
        polynomials.

        The recursive definition of **a single order** of the polynomial is

        .. math::

            H_{n+1}(x) = 2x H_{n}(x) - 2n H_{n-1}(x)

        with
        P_0 = 1
        P_1 = 2x

        Args:
            obs: The default space the PDF is defined in.
            coeffs: A list of the coefficients for the polynomials of order 1+ in the sum.
            apply_scaling: Rescale the data so that the actual limits represent (-1, 1).
            coeff0: The scaling factor of the 0th order polynomial. If not given, it is set to 1.
            extended: |@doc:pdf.init.extended| The overall yield of the PDF.
               If this is parameter-like, it will be used as the yield,
               the expected number of events, and the PDF will be extended.
               An extended PDF has additional functionality, such as the
               ``ext_*`` methods and the ``counts`` (for binned PDFs). |@docend:pdf.init.extended|
            norm: |@doc:pdf.init.norm| Normalization of the PDF.
               By default, this is the same as the default space of the PDF. |@docend:pdf.init.norm|
            name: Name of the polynomial
        """
        super().__init__(
            obs=obs,
            name=name,
            coeffs=coeffs,
            coeff0=coeff0,
            apply_scaling=apply_scaling,
            extended=extended,
            norm=norm,
        )

    def _poly_func(self, x):
        coeffs = convert_coeffs_dict_to_list(self.params)
        return hermite_shape(x=x, coeffs=coeffs)


class HermiteRepr(BaseRecursivePolynomialRepr):
    _implementation = Hermite
    hs3_type: Literal["Hermite"] = pydantic.Field("Hermite", alias="type")


def func_integral_hermite(limits, norm, params, model):
    lower, upper = limits.limit1d
    lower_rescaled = model._polynomials_rescale(lower)
    upper_rescaled = model._polynomials_rescale(upper)

    lower = z.convert_to_tensor(lower_rescaled)
    upper = z.convert_to_tensor(upper_rescaled)

    # the integral of hermite is a hermite_ni. We add the ni to the coeffs.
    coeffs = {"c_0": z.constant(0.0, dtype=model.dtype)}

    for name, coeff in params.items():
        ip1_coeff = int(name.split("_", 1)[-1]) + 1
        coeffs[f"c_{ip1_coeff}"] = coeff / z.convert_to_tensor(
            ip1_coeff * 2.0, dtype=model.dtype
        )
    coeffs = convert_coeffs_dict_to_list(coeffs)

    def indefinite_integral(limits):
        return hermite_shape(x=limits, coeffs=coeffs)

    integral = indefinite_integral(upper) - indefinite_integral(lower)
    integral = znp.reshape(integral, newshape=())
    integral *= 0.5 * model.space.area()  # rescale back to whole width

    return integral


hermite_limits_integral = Space(axes=0, limits=(Space.ANY_LOWER, Space.ANY_UPPER))
Hermite.register_analytic_integral(
    func=func_integral_hermite, limits=hermite_limits_integral
)


def convert_coeffs_dict_to_list(coeffs: Mapping) -> list:
    # HACK(Mayou36): how to solve elegantly? yield not a param, only a dependent?
    coeffs_list = []
    for i in range(len(coeffs)):
        try:
            coeffs_list.append(coeffs[f"c_{i}"])
        except (
            KeyError
        ):  # happens, if there are other parameters in there, such as a yield
            break
    return coeffs_list


# EOF



#  Copyright (c) 2023 zfit

from __future__ import annotations

from functools import partial
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import zfit

from collections.abc import Iterable

import tensorflow as tf

from ..core.interfaces import ZfitIndependentParameter, ZfitLoss
from ..util.exception import OperationNotAllowedError
from .baseminimizer import BaseStepMinimizer, minimize_supports


class WrapOptimizer(BaseStepMinimizer):
    def __init__(
        self,
        optimizer,
        tol=None,
        criterion=None,
        strategy=None,
        verbosity=None,
        name=None,
        **kwargs,
    ):
        """Wrap TensorFlow optimizers to have the zfit interface and behavior.

        .. note:: Different behavior of minimize

          While the ``minimize`` method in TensorFlow optimizers executes a single step of the minimization,
          the ``minimize`` method of a :class:`ZfitMinimizer` fully minimizes a function until convergence
          is reached. In order to execute a single step, use the :meth:`~WrapOptimizer.step` method (however,
          this is in general not necessary to do and rather inefficient).

        Args:
            optimizer: Instance of a :class:`tf.keras.optimizers.Optimizer`.
            All other arguments: these are passed through and have the same functionality as described
            in :py:class:`~BaseStepMinimizer`
        """
        if not isinstance(optimizer, tf.keras.optimizers.Optimizer):
            raise TypeError(
                f"optimizer {str(optimizer)} has to be from class Optimizer"
            )
        super().__init__(
            tol=tol,
            criterion=criterion,
            strategy=strategy,
            verbosity=verbosity,
            name=name,
            minimizer_options=None,
            **kwargs,
        )
        self._optimizer_tf = optimizer

    @minimize_supports(init=True)
    def _minimize(self, loss, params, init):
        try:
            return super()._minimize(loss, params, init)
        except ValueError as error:
            if "No gradients provided for any variable" in error.args[0]:
                raise OperationNotAllowedError(
                    "Cannot use TF optimizer with"
                    " a numerical gradient (non-TF function)"
                ) from None
            else:
                raise

    def _step(
        self,
        loss: ZfitLoss,
        params: Iterable[ZfitIndependentParameter],
        init: zfit.result.FitResult | None,
    ) -> tf.Tensor:
        # TODO(WrappedVariable): this is needed if we want to use wrapped Variables
        # import zfit
        # params = zfit.z.math._extract_tfparams(params)
        value = partial(loss.value, full=False)
        self._optimizer_tf.minimize(loss=value, var_list=params)
        return value()



#  Copyright (c) 2023 zfit
import pathlib

import matplotlib.pyplot as plt
import numpy as np
import pytest


@pytest.mark.parametrize("exact_nsample", [True, False], ids=["exact", "binomial sum"])
def test_yield_bias(exact_nsample, ntoys=300):
    import zfit
    import zfit.z.numpy as znp

    plot_folder = pathlib.Path(
        f"{'exact' if exact_nsample else 'binomial_sum'}_yield_sampling"
    )

    obs = zfit.Space("x", limits=(-10, 10))
    mu = zfit.Parameter("mu", 1.0, -4, 6)
    sigma = zfit.Parameter("sigma", 1.1, 0.1, 10)
    lambd = zfit.Parameter("lambda", -0.06, -1, -0.01)
    gauss = zfit.pdf.Gauss(mu=mu, sigma=sigma, obs=obs)
    exponential = zfit.pdf.Exponential(lambd, obs=obs)
    n_bkg = zfit.Parameter("n_bkg", 20000)
    n_sig = zfit.Parameter("n_sig", 1000)
    params = [mu, sigma, lambd, n_bkg, n_sig]
    gauss_extended = gauss.create_extended(n_sig)
    exp_extended = exponential.create_extended(n_bkg)
    model = zfit.pdf.SumPDF([gauss_extended, exp_extended])
    init_vals = [0.5, 1.2, -0.05, 20350, 2512]
    true_vals = [1.0, 1.1, -0.06, 20000, 3000]
    zfit.param.set_values(params, true_vals)
    true_nsig = true_vals[-1]
    gauss_sample = gauss.create_sampler(n=true_nsig)
    true_nbkg = true_vals[-2]
    exp_sample = exponential.create_sampler(n=true_nbkg)

    def sample_func():
        exp_sample.resample()
        gauss_sample.resample()
        gauss_val = gauss_sample.value()
        assert gauss_val.shape[0] == true_nsig
        exp_val = exp_sample.value()
        assert exp_val.shape[0] == true_nbkg
        return znp.concatenate([gauss_val, exp_val])

    data = model.create_sampler()
    data.sample_holder.assign(sample_func())
    nll = zfit.loss.ExtendedUnbinnedNLL(model=model, data=data)
    nsigs = []
    nbkgs = []
    minimizer = zfit.minimize.Minuit(gradient=False, tol=1e-05, mode=2)
    failures = 0
    for _ in range(ntoys):
        zfit.param.set_values(params, true_vals)
        if exact_nsample:
            data.sample_holder.assign(sample_func())
        else:
            data.resample(n=sum(true_vals[-2:]))
        for _ in range(10):
            zfit.param.set_values(params, init_vals)
            rnd_sig = np.random.uniform(-(true_nsig**0.5), true_nsig**0.5)
            rnd_bkg = np.random.uniform(-(true_nbkg**0.5), true_nbkg**0.5)
            rnd_sig *= (rnd_sig > 0) + 1
            rnd_sig += (rnd_sig > 0) - 0.5 + true_nsig**0.5 / 2
            rnd_bkg *= (rnd_bkg > 0) + 1
            rnd_bkg += (rnd_bkg > 0) - 0.5 + true_nbkg**0.5 / 2
            n_sig.set_value(true_nsig + rnd_sig)
            n_bkg.set_value(true_nbkg + rnd_bkg)
            result = minimizer.minimize(nll)
            if result.valid:
                break
        else:
            failures += 1
            continue
        nsig_res = float(result.params[n_sig]["value"])
        nsigs.append(nsig_res)
        nbkg_res = float(result.params[n_bkg]["value"])
        nbkgs.append(nbkg_res)
        assert nsig_res + nbkg_res == pytest.approx(true_nsig + true_nbkg, abs=0.6)
    nsigs_mean = np.mean(nsigs)
    std_nsigs_mean = np.std(nsigs) / ntoys**0.5 * 5
    nbkg_mean = np.mean(nbkgs)
    std_nbkg_mean = np.std(nbkgs) / ntoys**0.5 * 5
    plt.figure("yield_bias_toys")
    plt.title(
        f'{"Exact" if exact_nsample else "Binomial sum"} sampled. Fit with {minimizer.name}.'
    )

    counts, edges, _ = plt.hist(nsigs, bins=50, label=" Signal yields", alpha=0.5)
    npoints = 50
    plt.plot(
        np.ones(npoints) * true_nsig, np.linspace(0, np.max(counts)), "gx", label="true"
    )

    plt.plot(
        np.ones(npoints) * nsigs_mean,
        np.linspace(0, np.max(counts) * 2 / 3),
        "bo",
        label=f"Signal mean: {nsigs_mean:.2f} +- {std_nsigs_mean:.2f}",
    )

    plt.plot(
        np.ones(npoints) * nsigs_mean - np.std(nsigs),
        np.linspace(0, np.max(counts) * 0.2),
        "ro",
        label="-std",
    )

    plt.plot(
        np.ones(npoints) * nsigs_mean + np.std(nsigs),
        np.linspace(0, np.max(counts) * 0.2),
        "ro",
        label="+std",
    )

    plt.legend()
    pytest.zfit_savefig(folder=plot_folder)
    rel_err_sig = 0.001
    assert nsigs_mean == pytest.approx(true_nsig, rel=rel_err_sig, abs=std_nsigs_mean)
    rel_err_bkg = 0.001
    assert nbkg_mean == pytest.approx(true_nbkg, rel=rel_err_bkg, abs=std_nbkg_mean)



#  Copyright (c) 2023 zfit
import boost_histogram as bh
import hist
import numpy as np
import pytest
import tensorflow as tf


@pytest.fixture
def hist1():
    from zfit._data.binneddatav1 import BinnedData

    return BinnedData


@pytest.fixture
def holder1():
    from zfit.z import numpy as znp
    from zfit._data.binneddatav1 import BinnedHolder

    return BinnedHolder(
        tf.constant("asdf"), znp.random.uniform(size=[5]), znp.random.uniform(size=[5])
    )


@pytest.fixture
def holder2():
    from zfit.z import numpy as znp
    from zfit._data.binneddatav1 import BinnedHolder

    return BinnedHolder(
        tf.constant("asdf"), znp.random.uniform(size=[5]), znp.random.uniform(size=[5])
    )


@pytest.mark.skip(
    "Currently not a composite tensor, test when we can pass it through again"
)
def test_composite(holder1, holder2):
    from zfit.z import numpy as znp

    count = 0

    @tf.function
    def tot_variances(x):
        nonlocal count
        count += 1
        return znp.sum(x.variances)

    trueval = np.sum(holder1.variances)
    testval = tot_variances(holder1)
    assert pytest.approx(trueval, testval)
    actual_count = count
    _ = tot_variances(holder2)
    assert count == actual_count


def test_from_and_to_binned():
    h3 = hist.Hist(
        hist.axis.Regular(3, -3, 3, name="x", flow=False),
        hist.axis.Regular(2, -5, 5, name="y", flow=False),
        storage=hist.storage.Weight(),
    )

    x2 = np.random.randn(1_000)
    y2 = 0.5 * np.random.randn(1_000)

    h3.fill(x=x2, y=y2)

    from zfit._data.binneddatav1 import BinnedData

    h1 = BinnedData.from_hist(h3)
    for _ in range(10):  # make sure this works many times
        unbinned = h1.to_unbinned()
        binned = unbinned.to_binned(space=h1.space)
        np.testing.assert_allclose(binned.values(), h1.values())
        # we can't test the variances, this info is lost
        h1 = binned
    bh3 = bh.Histogram(h1)
    np.testing.assert_allclose(h1.values(), bh3.values())


def test_from_and_to_hist():
    h3 = hist.NamedHist(
        hist.axis.Regular(25, -3.5, 3, name="x", flow=False),
        hist.axis.Regular(21, -4, 5, name="y", flow=False),
        storage=hist.storage.Weight(),
    )

    x2 = np.random.randn(1_000)
    y2 = 0.5 * np.random.randn(1_000)

    h3.fill(x=x2, y=y2)

    from zfit._data.binneddatav1 import BinnedData

    for _ in range(10):  # make sure this works many times
        h1 = BinnedData.from_hist(h3)
        np.testing.assert_allclose(h1.variances(), h3.variances())
        np.testing.assert_allclose(h1.values(), h3.values())
        unbinned = h1.to_unbinned()
        assert unbinned.value().shape[1] == 2
        assert unbinned.value().shape[0] == unbinned.weights.shape[0]

        h3recreated = h1.to_hist()
        assert h3recreated == h3

    bh3 = bh.Histogram(h1)
    np.testing.assert_allclose(h1.variances(), bh3.variances())
    np.testing.assert_allclose(h1.values(), bh3.values())


def test_with_obs():
    from zfit._data.binneddatav1 import BinnedData

    h1 = hist.NamedHist(
        hist.axis.Regular(25, -3.5, 3, name="x", flow=False),
        hist.axis.Regular(21, -4, 5, name="y", flow=False),
        hist.axis.Regular(15, -2, 1, name="z", flow=False),
        storage=hist.storage.Weight(),
    )

    x2 = np.random.randn(1_000)
    y2 = 0.5 * np.random.randn(1_000)
    z2 = 0.3 * np.random.randn(1_000)

    h1.fill(x=x2, y=y2, z=z2)
    h = BinnedData.from_hist(h1)
    obs = ("x", "y", "z")
    obs2 = ("y", "x", "z")
    assert obs == h.obs
    h2 = h.with_obs(obs2)
    assert h2.obs == obs2
    np.testing.assert_allclose(h.values()[:, 3, 5], h2.values()[3, :, 5])
    np.testing.assert_allclose(h.variances()[:, 3, 5], h2.variances()[3, :, 5])


def test_valid_input():
    from zfit._data.binneddatav1 import BinnedHolder
    import zfit.z.numpy as znp
    import zfit
    from zfit.exception import ShapeIncompatibleError

    obs10 = zfit.Space("x", binning=zfit.binned.RegularBinning(10, -3, 4, name="x"))
    obs5 = zfit.Space("x", binning=zfit.binned.RegularBinning(5, -3, 4, name="x"))
    _ = BinnedHolder(
        obs10, znp.random.uniform(size=[10]), znp.random.uniform(size=[10])
    )
    with pytest.raises(tf.errors.InvalidArgumentError):
        _ = BinnedHolder(
            obs10, znp.random.uniform(size=[5]), znp.random.uniform(size=[5])
        )
    with pytest.raises(tf.errors.InvalidArgumentError):
        _ = BinnedHolder(
            obs5, znp.random.uniform(size=[10]), znp.random.uniform(size=[5])
        )
    with pytest.raises(tf.errors.InvalidArgumentError):
        _ = BinnedHolder(
            obs5, znp.random.uniform(size=[10]), znp.random.uniform(size=[10])
        )
    _ = BinnedHolder(obs5, znp.random.uniform(size=[5]), None)
    _ = BinnedHolder(obs10, znp.random.uniform(size=[10]), None)
    with pytest.raises(tf.errors.InvalidArgumentError):
        _ = BinnedHolder(obs5, znp.random.uniform(size=[10]), None)
    with pytest.raises(tf.errors.InvalidArgumentError):
        _ = BinnedHolder(obs10, znp.random.uniform(size=[5]), None)
    with pytest.raises(ShapeIncompatibleError):
        _ = BinnedHolder(
            obs10, znp.random.uniform(size=[10]), znp.random.uniform(size=[10, 2])
        )
    with pytest.raises(ShapeIncompatibleError):
        _ = BinnedHolder(
            obs10, znp.random.uniform(size=[10, 2]), znp.random.uniform(size=[10, 2])
        )
    with pytest.raises(ShapeIncompatibleError):
        _ = BinnedHolder(obs10, znp.random.uniform(size=[5, 2]), None)


def test_variance():
    import zfit
    import zfit.z.numpy as znp

    binning1 = zfit.binned.RegularBinning(3, -3.5, 3, name="x")
    obs = zfit.Space("x", binning=binning1)
    values = znp.array([100.0, 200, 50])
    data = zfit.data.BinnedData.from_tensor(obs, values=values, variances=True)
    data2 = zfit.data.BinnedData.from_tensor(
        obs, values=values, variances=values**0.5
    )
    np.testing.assert_allclose(data.variances(), data2.variances())



#  Copyright (c) 2023 zfit
import pytest

import zfit
from zfit import z


@pytest.mark.skipif(
    zfit.run.get_graph_mode() is False,
    reason="If not using graph mode, we cannot test if graphs work or not.",
)
@pytest.mark.parametrize("cachesize", [1, 10, 1000, None])
def test_modes(cachesize):
    counts = 0

    zfit.run.set_graph_cache_size(cachesize)
    if cachesize is not None:
        assert all(
            register.cachesize == cachesize
            for register in zfit.z.zextension.FunctionWrapperRegistry.registries
        )

    @z.function(wraps="42")
    def func(x):
        nonlocal counts
        counts += 1
        return z.random.get_prng().uniform(shape=()) * x

    func(5)

    counts_after_compile = (
        counts  # TF can run through a function n times when compiling
    )
    func(5)
    assert counts == counts_after_compile
    func(5)
    assert counts == counts_after_compile
    func(5)
    assert counts == counts_after_compile
    zfit.run.set_graph_mode(False)
    func(5)
    assert counts == counts_after_compile + 1
    zfit.run.set_mode_default()
    func(5)
    func(5)
    func(5)
    assert (
        counts == counts_after_compile + 1
    )  # No compilation occurred, default is graph mode
    zfit.run.clear_graph_cache()
    counts = 0
    func(5)
    counts_after_compile = counts
    func(5)
    assert counts == counts_after_compile
    func(5)
    assert counts == counts_after_compile

    # use with everywhere in other to reset it correctly
    with zfit.run.set_autograd_mode(False):
        assert zfit.settings.options.numerical_grad
        with zfit.run.set_mode_default():
            assert not zfit.settings.options.numerical_grad

            with zfit.run.set_graph_mode(True):
                assert zfit.run.get_graph_mode()
                with zfit.run.set_graph_mode(False):
                    assert not zfit.run.get_graph_mode()
                assert zfit.run.get_graph_mode()

            with zfit.run.set_autograd_mode(True):
                assert zfit.run.get_autograd_mode()
                with zfit.run.set_autograd_mode(False):
                    assert not zfit.run.get_autograd_mode()
                assert zfit.run.get_autograd_mode()



#  Copyright (c) 2022 zfit

import numpy as np

import zfit
from zfit.core.binning import histogramdd

data1 = np.random.normal(size=(1000, 3))
obs1 = zfit.Space("obs1", limits=(-100, 300))
obs2 = zfit.Space("obs2", limits=(-100, 300))
obs3 = zfit.Space("obs3", limits=(-100, 300))

obs = obs1 * obs2 * obs3


def test_histogramdd():
    histdd_kwargs = {"sample": data1}
    hist = histogramdd(**histdd_kwargs)
    bincount_np, edges_np = zfit.run(hist)
    bincount_true, edges_true = np.histogramdd(**histdd_kwargs)
    np.testing.assert_allclose(bincount_true, bincount_np)
    np.testing.assert_allclose(edges_true, edges_np)



#  Copyright (c) 2023 zfit

from __future__ import annotations

from typing import TYPE_CHECKING, List

if TYPE_CHECKING:
    import zfit

from collections.abc import Iterable
from collections.abc import Callable

import numdifftools
import tensorflow as tf
import tensorflow_probability as tfp

from ..util.container import convert_to_container
from ..util.deprecation import deprecated
from . import numpy as znp
from .tools import _auto_upcast
from .zextension import convert_to_tensor


def poly_complex(*args, real_x=False):
    """Complex polynomial with the last arg being x.

    Args:
        *args: Coefficients of the polynomial
        real_x: If True, x is assumed to be real.

    Returns:
    """
    from .. import z

    args = list(args)
    x = args.pop()
    if real_x is not None:
        pow_func = znp.power
    else:
        pow_func = z.nth_pow
    return tf.add_n(
        [coef * z.to_complex(pow_func(x, p)) for p, coef in enumerate(args)]
    )


def numerical_gradient(func: Callable, params: Iterable["zfit.Parameter"]) -> tf.Tensor:
    """Calculate numerically the gradient of func() with respect to ``params``.

    Args:
        func: Function without arguments that depends on ``params``
        params: Parameters that ``func`` implicitly depends on and with respect to which the
            derivatives will be taken.

    Returns:
        Gradients
    """
    from ..core.parameter import assign_values

    params = convert_to_container(params)

    def wrapped_func(param_values):
        assign_values(params, param_values)
        value = func()
        if hasattr(value, "numpy"):
            value = value.numpy()
        return value

    param_vals = znp.stack(params)
    original_vals = [param.value() for param in params]
    grad_func = numdifftools.Gradient(wrapped_func, order=2, base_step=1e-4)
    if tf.executing_eagerly():
        grad_vals = grad_func(param_vals)
        gradient = convert_to_tensor(grad_vals)
    else:
        gradient = tf.numpy_function(grad_func, inp=[param_vals], Tout=tf.float64)
    if gradient.shape == ():
        gradient = znp.reshape(gradient, newshape=(1,))
    gradient.set_shape(param_vals.shape)
    assign_values(params, original_vals)
    return gradient


def numerical_value_gradient(
    func: Callable, params: Iterable[zfit.Parameter]
) -> [tf.Tensor, tf.Tensor]:
    """Calculate numerically the gradients of ``func()`` with respect to ``params``, also returns the value of
    ``func()``.

    Args:
        func: Function without arguments that depends on ``params``
        params: Parameters that ``func`` implicitly depends on and with respect to which the
            derivatives will be taken.

    Returns:
        Value, gradient
    """
    return func(), numerical_gradient(func, params)


deprecated(None, "Use `numerical_value_gradient` instead.")


def numerical_value_gradients(*args, **kwargs):
    return numerical_value_gradients(*args, **kwargs)


def numerical_hessian(
    func: Callable | None, params: Iterable[zfit.Parameter], hessian=None
) -> tf.Tensor:
    """Calculate numerically the hessian matrix of func with respect to ``params``.

    Args:
        func: Function without arguments that depends on ``params``
        params: Parameters that ``func`` implicitly depends on and with respect to which the
            derivatives will be taken.

    Returns:
        Hessian matrix
    """
    from ..core.parameter import assign_values

    params = convert_to_container(params)

    def wrapped_func(param_values):
        assign_values(params, param_values)
        value = func()
        if hasattr(value, "numpy"):
            value = value.numpy()
        return value

    param_vals = znp.stack(params)
    original_vals = [param.value() for param in params]

    if hessian == "diag":
        hesse_func = numdifftools.Hessdiag(
            wrapped_func,
            order=2,
            # TODO: maybe add step to remove numerical problems?
            base_step=1e-4,
        )
    else:
        hesse_func = numdifftools.Hessian(
            wrapped_func,
            order=2,
            base_step=1e-4,
        )
    if tf.executing_eagerly():
        computed_hessian = convert_to_tensor(hesse_func(param_vals))
    else:
        computed_hessian = tf.numpy_function(
            hesse_func, inp=[param_vals], Tout=tf.float64
        )
    n_params = param_vals.shape[0]
    if hessian == "diag":
        computed_hessian.set_shape((n_params,))
    else:
        computed_hessian.set_shape((n_params, n_params))

    assign_values(params, original_vals)
    return computed_hessian


def numerical_value_gradient_hessian(
    func: Callable | None,
    params: Iterable[zfit.Parameter],
    gradient: Callable | None = None,
    hessian: str | None = None,
) -> [tf.Tensor, tf.Tensor, tf.Tensor]:
    """Calculate numerically the gradients and hessian matrix of ``func()`` wrt ``params``; also return ``func()``.

    Args:
        func: Function without arguments that depends on ``params``
        params: Parameters that ``func`` implicitly depends on and with respect to which the
            derivatives will be taken.

    Returns:
        Value, gradient and hessian matrix
    """
    if params is None:
        raise ValueError("params cannot be None")
    if func is None and gradient is None:
        raise ValueError("Either func or grad has to be given")
    value, gradients = numerical_value_gradient(func, params)
    hessian = numerical_hessian(func, params, hessian=hessian)

    return value, gradients, hessian


@deprecated(None, "Use `numerical_value_gradient_hessian` instead.")
def numerical_value_gradients_hessian(*args, **kwargs):
    return numerical_value_gradient_hessian(*args, **kwargs)


def autodiff_gradient(func: Callable, params: Iterable[zfit.Parameter]) -> tf.Tensor:
    """Calculate using autodiff the gradients of ``func()`` wrt ``params``.

    Automatic differentiation (autodiff) is a way of retreiving the derivative of x wrt y. It works by consecutively
    applying the chain rule. All that is needed is that every operation knows its own derivative.
    TensorFlow implements this and anything using ``tf.*`` operations only can use this technique.

        Args:
            func: Function without arguments that depends on ``params``
            params: Parameters that ``func`` implicitly depends on and with respect to which the
                derivatives will be taken.

        Returns:
            Gradient
    """
    return autodiff_value_gradients(func, params)[1]


def _extract_tfparams(
    params: Iterable[zfit.Parameter] | zfit.Parameter,
) -> List[tf.Variable]:
    """Extract the tf.Variable from the parameters.

    Args:
        params:

    Returns:
        tf.Variable
    """
    return params
    # TODO(WrappedVariable): this is needed if we want to use wrapped Variables
    # import zfit
    # params = convert_to_container(params)
    # tf_params = []
    # for param in params:
    #     if isinstance(param, tf.Variable):
    #
    #         # TODO: reactivate if WrappedVariables are used
    #         # if isinstance(param, zfit.Parameter):
    #         #     raise ValueError("The parameter cannot be a tf.Variable and a zfit.Parameter at the same time.")
    #         variable = param
    #     else:
    #         if not isinstance(param, zfit.Parameter):
    #             raise ValueError("The parameter has to be either a tf.Variable or a zfit.Parameter.")
    #         variable = param.variable
    #     tf_params.append(variable)
    # return tf_params


def autodiff_value_gradient(
    func: Callable, params: Iterable[zfit.Parameter]
) -> [tf.Tensor, tf.Tensor]:
    """Calculate using autodiff the gradients of ``func()`` wrt ``params``; also return ``func()``.

    Automatic differentiation (autodiff) is a way of retreiving the derivative of x wrt y. It works by consecutively
    applying the chain rule. All that is needed is that every operation knows its own derivative.
    TensorFlow implements this and anything using ``tf.*`` operations only can use this technique.

        Args:
            func: Function without arguments that depends on ``params``
            params: Parameters that ``func`` implicitly depends on and with respect to which the
                derivatives will be taken.

        Returns:
            Value and gradient
    """
    # TODO(WrappedVariable): this is needed if we want to use wrapped Variables
    # params = _extract_tfparams(params)
    with tf.GradientTape(
        persistent=False,  # needs to be persistent for a call from hessian.
        watch_accessed_variables=False,
    ) as tape:
        tape.watch(params)
        value = func()
    gradients = tape.gradient(value, sources=params)
    gradients = znp.asarray(gradients)
    return value, gradients


@deprecated(None, "Use `autodiff_value_gradient` instead.")
def autodiff_value_gradients(*args, **kwargs):
    return autodiff_value_gradient(*args, **kwargs)


def autodiff_hessian(
    func: Callable, params: Iterable[zfit.Parameter], hessian=None
) -> tf.Tensor:
    """Calculate using autodiff the hessian matrix of ``func()`` wrt ``params``.

    Automatic differentiation (autodiff) is a way of retrieving the derivative of x wrt y. It works by consecutively
    applying the chain rule. All that is needed is that every operation knows its own derivative.
    TensorFlow implements this and anything using ``tf.*`` operations only can use this technique.

        Args:
            func: Function without arguments that depends on ``params``
            params: Parameters that ``func`` implicitly depends on and with respect to which the
                derivatives will be taken.

        Returns:
            Hessian matrix
    """

    return automatic_value_gradients_hessian(func, params, hessian=hessian)[2]


def automatic_value_gradient_hessian(
    func: Callable = None,
    params: Iterable[zfit.Parameter] = None,
    value_grad_func=None,
    hessian=None,
) -> [tf.Tensor, tf.Tensor, tf.Tensor]:
    """Calculate using autodiff the gradients and hessian matrix of ``func()`` wrt ``params``; also return ``func()``.

    Automatic differentiation (autodiff) is a way of retreiving the derivative of x wrt y. It works by consecutively
    applying the chain rule. All that is needed is that every operation knows its own derivative.
    TensorFlow implements this and anything using ``tf.*`` operations only can use this technique.

        Args:
            func: Function without arguments that depends on ``params``
            params: Parameters that ``func`` implicitly depends on and with respect to which the
                derivatives will be taken.

        Returns:
            Value, gradient and hessian matrix
    """
    if params is None:
        raise ValueError("Parameters have to be specified, are currently None.")
    if func is None and value_grad_func is None:
        ValueError("Either `func` or `value_grad_func` has to be specified.")

    from .. import z

    persistant = (
        hessian == "diag" or tf.executing_eagerly()
    )  # currently needed, TODO: can we better parallelize that?
    # TODO(WrappedVariable): this is needed if we want to use wrapped Variables
    # params = _extract_tfparams(params)
    with tf.GradientTape(persistent=persistant, watch_accessed_variables=False) as tape:
        tape.watch(params)
        if callable(value_grad_func):
            loss, gradients = value_grad_func(params)
        else:
            loss, gradients = autodiff_value_gradients(func=func, params=params)
        if hessian == "diag":
            gradients = tf.unstack(gradients)
            # gradients_tf = znp.stack(gradients)
    if hessian == "diag":
        computed_hessian = znp.stack(
            [
                tape.gradient(grad, sources=param)
                for param, grad in zip(params, gradients)
            ]
        )
        # gradfunc = lambda par_grad: tape.gradient(par_grad[0], sources=par_grad[1])
        # computed_hessian = tf.vectorized_map(gradfunc, zip(params, gradients))
    else:
        computed_hessian = z.convert_to_tensor(
            tape.jacobian(
                gradients,
                sources=params,
                experimental_use_pfor=True,  # causes TF bug? Slow..
            )
        )
    del tape
    return loss, gradients, computed_hessian


@deprecated(None, "Use `automatic_value_gradient_hessian` instead.")
def automatic_value_gradients_hessian(*args, **kwargs):
    return automatic_value_gradient_hessian(*args, **kwargs)


# @z.function  # TODO: circular import, improve?
def reduce_geometric_mean(input_tensor, axis=None, weights=None, keepdims=False):
    if weights is not None:
        log_mean = tf.nn.weighted_moments(
            log(input_tensor), axes=axis, frequency_weights=weights
        )[0]
    else:
        log_mean = znp.mean(znp.log(input_tensor), axis=axis, keepdims=keepdims)
    return znp.exp(log_mean)


def log(x):
    x = _auto_upcast(x)
    return _auto_upcast(znp.log(x=x))


def weighted_quantile(x, quantiles, weights=None, side="middle"):
    """Very close to numpy.percentile, but supports weights.

    NOTE: quantiles should be in [0, 1]!
    :param x: tensor with data
    :param quantiles: array-like with many quantiles needed
    :param weights: array-like of the same length as ``x``
    :return: numpy.array with computed quantiles.
    """
    if weights is None:
        return tfp.stats.percentile(x, 100 * quantiles)
    x = znp.array(x)
    quantiles = znp.array(quantiles)
    quantiles = znp.reshape(quantiles, (-1,))
    weights = znp.array(weights)

    sorter = znp.argsort(x)
    x = tf.gather(x, sorter)
    weights = tf.gather(weights, sorter)

    weighted_quantiles = znp.cumsum(weights) - 0.5 * weights

    weighted_quantiles /= znp.sum(weights)
    if side == "middle":
        quantile_index_left = tf.searchsorted(
            weighted_quantiles, quantiles, side="left"
        )
        quantile_index_right = tf.searchsorted(
            weighted_quantiles, quantiles, side="right"
        )

        calculated_left = tf.gather(x, quantile_index_left)
        calculated_right = tf.gather(x, quantile_index_right)
        calculated = (calculated_left + calculated_right) / 2
    elif side in ("left", "right"):
        quantile_index = tf.searchsorted(weighted_quantiles, quantiles, side=side)

        calculated = tf.gather(x, quantile_index)
    return calculated



#  Copyright (c) 2023 zfit

import matplotlib.pyplot as plt
import mplhep
import numpy as np

import zfit

plt.style.use(mplhep.style.LHCb2)

n_bins = 50

# create space
obs_binned = zfit.Space("x", binning=zfit.binned.RegularBinning(50, -10, 10, name="x"))
obs = obs_binned.with_binning(None)

# parameters
mu = zfit.Parameter("mu", 1.0, -4, 6)
sigma = zfit.Parameter("sigma", 1.0, 0.1, 10)
lambd = zfit.Parameter("lambda", -0.06, -1, -0.01)

# model building, pdf creation
gauss = zfit.pdf.Gauss(mu=mu, sigma=sigma, obs=obs)
exponential = zfit.pdf.Exponential(lambd, obs=obs)

n_bkg = zfit.Parameter("n_bkg", 5000 * (1 - 0.3))
n_sig = zfit.Parameter("n_sig", 5000 * 0.3)
gauss_extended = gauss.create_extended(n_sig)
exp_extended = exponential.create_extended(n_bkg)
model_unbinned = zfit.pdf.SumPDF([gauss_extended, exp_extended])

# make binned
model = zfit.pdf.BinnedFromUnbinnedPDF(model_unbinned, space=obs_binned)

# data
n_sample = 5000
data = model.sample(n=n_sample)

n_bins = 50

plot_scaling = n_sample / n_bins * obs.area()

x = np.linspace(-10, 10, 1000)


def plot_pdf(title):
    plt.figure()
    # plt.title(title)
    y = model.pdf(x).numpy()
    y_gauss = (gauss.pdf(x) * model_unbinned.params["frac_0"]).numpy()
    y_exp = (exponential.pdf(x) * model_unbinned.params["frac_1"]).numpy()
    plt.plot(x, y * plot_scaling, label="Sum - Binned Model")
    plt.plot(x, y_gauss * plot_scaling, label="Gauss - Signal")
    plt.plot(x, y_exp * plot_scaling, label="Exp - Background")
    # mplhep.histplot(np.histogram(data_np, bins=n_bins), yerr=True, color='black', histtype='errorbar')
    h = data.to_hist()
    mplhep.histplot(h, yerr=h.counts() ** 0.5, color="black", histtype="errorbar")
    plt.ylabel("Counts")
    plt.xlabel("obs: $B_{mass}$")
    plt.legend()
    plt.savefig(title + ".pdf")


# set the values to a start value for the fit
mu.set_value(0.5)
sigma.set_value(1.2)
lambd.set_value(-0.05)

# create NLL
nll = zfit.loss.ExtendedBinnedNLL(model=model, data=data)

# create a minimizer
minimizer = zfit.minimize.Minuit()

plot_pdf("before fit")

result = minimizer.minimize(nll)
print(result.params)
# do the error calculations, here with hesse, than with minos
param_hesse = result.hesse()
(
    param_errors,
    _,
) = result.errors()  # this returns a new FitResult if a new minimum was found
print(result.valid)  # check if the result is still valid

# plot the data

plot_pdf("after fit")
# uncomment to display plots
plt.show()



#  Copyright (c) 2022 zfit


class Singleton:
    __instance = None

    def __new__(cls, *args, **kwargs):
        instance = cls.__instance
        if instance is None:
            instance = super().__new__(cls)
            cls.__instance = instance

        return instance


class NotSpecified(Singleton):
    def __bool__(self):
        return False

    def __repr__(self):
        return "None"


NONE = NotSpecified()


class ZfitNotImplemented:
    def __new__(cls, *args, **kwargs):
        raise RuntimeError(
            "Cannot create an instance of it, meant to be used as a single object reference."
        )


class RuntimeDependency:
    def __init__(self, name, how=None):
        if how is None:
            how = ""
        self.__name = name
        self.__how = how

    def __getattr__(self, item):
        if item in self.__dict__:
            return self.__dict__[item]
        raise ImportError(
            f"This requires {self.__name}"
            " to be installed. You can usually install it with"
            f"`pip install zfit[{self.__name}]` or"
            f"`pip install zfit[all]`."
            f"{self.__how}"
        )



#  Copyright (c) 2022 zfit

import tensorflow as tf

from ..settings import ztypes


def generate_1d_grid(
    data, num_grid_points, absolute_boundary=0.0, relative_boundary=0.05
):
    minimum = tf.math.reduce_min(data)
    maximum = tf.math.reduce_max(data)
    space_width = maximum - minimum
    outside_borders = tf.maximum(relative_boundary * space_width, absolute_boundary)

    return tf.linspace(
        minimum - outside_borders, maximum + outside_borders, num=num_grid_points
    )


def bin_1d(binning_method, data, grid, weights=None):
    if binning_method == "simple":
        return bin_1d_simple(data, grid, weights)
    elif binning_method == "linear":
        return bin_1d_linear(data, grid, weights)
    else:
        raise ValueError(
            f"Binning method '{binning_method}' not supported, only 'simple' or 'linear'."
        )


def bin_1d_simple(data, grid, weights=None):
    if weights is None:
        bincount = tf.cast(
            tf.histogram_fixed_width(
                data,
                [tf.math.reduce_min(grid), tf.math.reduce_max(grid)],
                tf.size(grid),
            ),
            ztypes.float,
        )
    else:
        bincount = _bin_1d_weighted(data, grid, weights, "simple")

    return bincount


def bin_1d_linear(data, grid, weights=None):
    return _bin_1d_weighted(data, grid, weights)


def _bin_1d_weighted(data, grid, weights, method="linear"):
    if weights is None:
        weights = tf.ones_like(data, ztypes.float)

    weights = weights / tf.reduce_sum(weights)

    grid_size = tf.size(grid)
    grid_min = tf.math.reduce_min(grid)
    grid_max = tf.math.reduce_max(grid)
    num_intervals = tf.math.subtract(grid_size, tf.constant(1))
    dx = tf.math.divide(
        tf.math.subtract(grid_max, grid_min), tf.cast(num_intervals, ztypes.float)
    )

    transformed_data = tf.math.divide(tf.math.subtract(data, grid_min), dx)

    # Compute the integral and fractional part of the data
    # The integral part is used for lookups, the fractional part is used
    # to weight the data
    integral = tf.math.floor(transformed_data)
    fractional = tf.math.subtract(transformed_data, integral)

    if method == "simple":
        fractional = tf.cast(fractional > 0.5, fractional.dtype) * fractional

    # Compute the weights for left and right side of the linear binning routine
    frac_weights = tf.math.multiply(fractional, weights)
    neg_frac_weights = tf.math.subtract(weights, frac_weights)

    # tf.math.bincount only works with tf.int32
    bincount_left = tf.roll(
        tf.concat(
            tf.math.bincount(
                tf.cast(integral, tf.int32),
                weights=frac_weights,
                minlength=grid_size,
                maxlength=grid_size,
            ),
            tf.constant(0),
        ),
        shift=1,
        axis=0,
    )
    bincount_right = tf.math.bincount(
        tf.cast(integral, tf.int32),
        weights=neg_frac_weights,
        minlength=grid_size,
        maxlength=grid_size,
    )

    bincount = tf.cast(tf.add(bincount_left, bincount_right), ztypes.float)

    return bincount



#  Copyright (c) 2023 zfit

from __future__ import annotations

from collections.abc import Iterable
from functools import wraps
from typing import Any

import tensorflow as tf
import tensorflow_probability as tfp

from .zextension import function as function

__all__ = ["counts_multinomial", "sample_with_replacement"]

from ..settings import ztypes
from ..z import numpy as znp

generator = None


def get_prng():
    """Get the global random number generator.

    Returns:
        zfit random number generator
    """
    global generator
    if generator is None:  # initialization
        generator = tf.random.Generator.from_non_deterministic_state()
    return generator


def sample_with_replacement(
    a: tf.Tensor, axis: int, sample_shape: tuple[int]
) -> tf.Tensor:
    """Sample from ``a`` with replacement to return a Tensor with ``sample_shape``.

    Args:
        a (): Tensor to sample from
        axis (): int axis to sample along
        sample_shape (): Shape of the new samples along the axis

    Returns:
        tf.Tensor of shape a.shape[:axis] + samples_shape + a.shape[axis + 1:]


    Examples:

    .. code:: pycon

        >>> a = tf.random.uniform(shape=(10, 20, 30), dtype=tf.float32)
        >>> random_choice(a, axis=0)
        <tf.Tensor 'GatherV2:0' shape=(1, 20, 30) dtype=float32>
        >>> random_choice(a, axis=1, samples_shape=(2, 3))
        <tf.Tensor 'GatherV2_2:0' shape=(10, 2, 3, 30) dtype=float32
        >>> random_choice(a, axis=0, samples_shape=(100,))
        <tf.Tensor 'GatherV2_3:0' shape=(100, 20, 30) dtype=float32>
    """

    dim = tf.shape(a)[axis]
    choice_indices = get_prng().uniform(
        sample_shape, minval=0, maxval=dim, dtype=tf.int32
    )
    samples = tf.gather(a, choice_indices, axis=axis)
    return samples


def counts_multinomial(
    total_count: int | tf.Tensor,
    probs: Iterable[float | tf.Tensor] = None,
    logits: Iterable[float | tf.Tensor] = None,
    dtype=tf.int32,
) -> tf.Tensor:
    """Get the number of counts for different classes with given probs/logits.

    Args:
        total_count: The total number of draws.
        probs: Length k (number of classes) object where the k-1th entry contains the probability to
            get a single draw from the class k. Have to be from [0, 1] and sum up to 1.
        logits: Same as probs but from [-inf, inf] (will be transformet to [0, 1])

    Returns:
        Shape (k,) tensor containing the number of draws.
    """
    from .. import z

    total_count = tf.convert_to_tensor(total_count)
    probs = z.convert_to_tensor(probs) if probs is not None else probs
    logits = tf.convert_to_tensor(logits) if logits is not None else logits

    if probs is not None:
        probs = tf.cast(probs, dtype=tf.float64)
        float_dtype = probs.dtype
    elif logits is not None:
        logits = tf.cast(logits, tf.float64)
        float_dtype = logits.dtype
    else:
        raise ValueError("Exactly one of `probs` or`logits` have to be specified")
    total_count = tf.cast(total_count, dtype=float_dtype)

    # needed since otherwise shape of sample will be (1, n_probs)
    # total_count = tf.broadcast_to(total_count, shape=probs_logits_shape)

    # @function

    return _wrapped_multinomial_func(dtype, logits, probs, total_count)


@function(wraps="tensor")
def _wrapped_multinomial_func(dtype, logits, probs, total_count):
    if probs is not None:
        shape = tf.shape(probs)
        probs = znp.reshape(probs, [-1])
    else:
        shape = tf.shape(logits)
        logits = znp.reshape(logits, [-1])
    dist = tfp.distributions.Multinomial(
        total_count=total_count, probs=probs, logits=logits
    )
    counts_flat = dist.sample()
    counts_flat = tf.cast(counts_flat, dtype=dtype)
    counts = znp.reshape(counts_flat, shape)
    return counts


@wraps(tf.random.normal)
def normal(shape, mean=0.0, stddev=1.0, dtype=ztypes.float, name=None):
    return get_prng().normal(
        shape=shape, mean=mean, stddev=stddev, dtype=dtype, name=name
    )


@wraps(tf.random.uniform)
def uniform(shape, minval=0, maxval=None, dtype=ztypes.float, name=None):
    return get_prng().uniform(
        shape=shape, minval=minval, maxval=maxval, dtype=dtype, name=name
    )


@wraps(tf.random.poisson)
def poisson(
    lam: Any,
    shape: Any,
    seed: Any = None,
    dtype: tf.DType = ztypes.float,
    name: Any = None,
):
    if seed is None:
        seed = get_prng().make_seeds(1)[:, 0]
    return tf.random.stateless_poisson(
        lam=lam, seed=seed, shape=shape, dtype=dtype, name=name
    )


def shuffle(value, seed=None, name=None):
    if seed is None:
        seed = get_prng().make_seeds(1)[:, 0]
    return tf.random.experimental.stateless_shuffle(value, seed=seed, name=name)



#  Copyright (c) 2022 zfit

from .core.parameter import (
    ComplexParameter,
    ComposedParameter,
    ConstantParameter,
    Parameter,
    assign_values,
    convert_to_parameter,
    convert_to_parameters,
    set_values,
)

__all__ = [
    "ConstantParameter",
    "Parameter",
    "ComposedParameter",
    "ComplexParameter",
    "convert_to_parameter",
    "convert_to_parameters",
    "set_values",
    "assign_values",
]



#  Copyright (c) 2023 zfit
import numpy as np

import zfit


# Set up the simple model for the test
def create_loss(mu, sigma):
    obs = zfit.Space("x", limits=(-10, 10))
    gauss = zfit.pdf.Gauss(mu=mu, sigma=sigma, obs=obs)

    data_np = np.random.normal(mu.value(), sigma.value(), size=5)
    data_np = zfit.Data.from_numpy(obs=obs, array=data_np)
    data = gauss.sample(n=5)
    return zfit.loss.UnbinnedNLL(model=[gauss, gauss], data=[data, data_np])


# Test the reproducibility of the fit
def test_reproducibility():
    # Create model and data

    mu = zfit.Parameter("mu", 0.0, -5, 5)
    sigma = zfit.Parameter("sigma", 1.0, 0.1, 10)

    # Set up minimizer
    minimizer = zfit.minimize.Minuit()
    params = [mu, sigma]
    vals = [0.2, 0.7]
    # Perform fit with fixed seed
    seed = 4294
    zfit.settings.set_seed(seed)
    loss = create_loss(mu=mu, sigma=sigma)
    with zfit.param.set_values(params, vals):
        result_fixed_seed1 = minimizer.minimize(loss)

    # Perform fit again with the same fixed seed
    zfit.settings.set_seed(seed)
    loss = create_loss(mu=mu, sigma=sigma)
    with zfit.param.set_values(params, vals):
        result_fixed_seed2 = minimizer.minimize(loss)

    # Perform fit without a fixed seed (set to None)
    zfit.settings.set_seed(None)
    loss = create_loss(mu=mu, sigma=sigma)
    with zfit.param.set_values(params, vals):
        result_no_seed = minimizer.minimize(loss)

    zfit.settings.set_seed(seed)
    loss = create_loss(mu=mu, sigma=sigma)
    with zfit.param.set_values(params, vals):
        result_fixed_seed3 = minimizer.minimize(loss)

    zfit.settings.set_seed(None)
    loss = create_loss(mu=mu, sigma=sigma)
    with zfit.param.set_values(params, vals):
        result_no_seed2 = minimizer.minimize(loss)

    # Check if the results with fixed seeds are the same
    assert np.allclose(
        result_fixed_seed1.values, result_fixed_seed2.values
    ), "Results with the same fixed seed should be the same"

    # Check if the results without a fixed seed are different
    assert not np.allclose(
        result_fixed_seed1.values, result_no_seed.values
    ), "Results without a fixed seed should be different"

    assert np.allclose(
        result_fixed_seed3.values, result_fixed_seed2.values
    ), "Results with the same fixed seed should be the same"
    assert not np.allclose(
        result_fixed_seed3.values, result_no_seed2.values
    ), "Results without a fixed seed should be different"
    assert not np.allclose(
        result_no_seed.values, result_no_seed2.values
    ), "Results without a fixed seed should be different"

    # check if numpy and tensorflow seeds can be set separately
    zfit.settings.set_seed(seed, numpy=2)
    loss = create_loss(mu=mu, sigma=sigma)
    with zfit.param.set_values(params, vals):
        result_fixed_seed_npvary = minimizer.minimize(loss)

    assert not np.allclose(
        result_fixed_seed_npvary.values, result_fixed_seed2.values
    ), "Results with the same fixed seed should be the same"
    assert not np.allclose(
        result_fixed_seed_npvary.values, result_no_seed.values
    ), "Results without a fixed seed should be different"

    zfit.settings.set_seed(seed, backend=3)
    loss = create_loss(mu=mu, sigma=sigma)
    with zfit.param.set_values(params, vals):
        result_fixed_seed_tvary = minimizer.minimize(loss)

    # TODO: that means that the rnd does not depend on the global? or what?
    # assert not np.allclose(result_fixed_seed_tvary.values, result_fixed_seed2.values), "Results with the same fixed seed should be the same"
    assert not np.allclose(
        result_fixed_seed_tvary.values, result_no_seed.values
    ), "Results without a fixed seed should be different"

    zfit.settings.set_seed(seed, numpy=2)
    loss = create_loss(mu=mu, sigma=sigma)
    with zfit.param.set_values(params, vals):
        result_fixed_seed_npvary2 = minimizer.minimize(loss)

    assert np.allclose(
        result_fixed_seed_npvary2.values, result_fixed_seed_npvary.values
    ), "Results with the same fixed seed should be the same"
    assert not np.allclose(
        result_fixed_seed_npvary2.values, result_no_seed.values
    ), "Results without a fixed seed should be different"

    zfit.settings.set_seed(seed, backend=3)
    loss = create_loss(mu=mu, sigma=sigma)
    with zfit.param.set_values(params, vals):
        result_fixed_seed_tvary2 = minimizer.minimize(loss)

    assert np.allclose(
        result_fixed_seed_tvary2.values, result_fixed_seed_tvary.values
    ), "Results with the same fixed seed should be the same"
    assert not np.allclose(
        result_fixed_seed_tvary2.values, result_no_seed.values
    ), "Results without a fixed seed should be different"

    zfit.settings.set_seed(seed, numpy=4, backend=5)
    loss = create_loss(mu=mu, sigma=sigma)
    with zfit.param.set_values(params, vals):
        result_fixed_seed_tvary_np4 = minimizer.minimize(loss)

    assert not np.allclose(
        result_fixed_seed_tvary_np4.values, result_fixed_seed_tvary.values
    ), "Results with the same fixed seed should be the same"
    assert not np.allclose(
        result_fixed_seed_tvary_np4.values, result_no_seed.values
    ), "Results without a fixed seed should be different"

    zfit.settings.set_seed(seed, numpy=4, backend=5)
    loss = create_loss(mu=mu, sigma=sigma)
    with zfit.param.set_values(params, vals):
        result_fixed_seed_tvary_np4_2 = minimizer.minimize(loss)

    assert np.allclose(
        result_fixed_seed_tvary_np4_2.values, result_fixed_seed_tvary_np4.values
    ), "Results with the same fixed seed should be the same"



#  Copyright (c) 2023 zfit
from typing import (
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Tuple,
    TypeVar,
    Union,
    Mapping,
)

import numpy as np
import tensorflow as tf

# space
from ordered_set import OrderedSet

# The #: symbols at the end of every type alias are for marking the module level variables
# as documented, such that sphinx will document them.
from uhi.typing.plottable import PlottableHistogram

LowerTypeInput = Union[tf.Tensor, np.ndarray, Tuple[float], List[float], float]  #:
LowerTypeReturn = Union[np.ndarray, tf.Tensor, None, bool]  #:

UpperTypeInput = LowerTypeInput  #:

UpperTypeReturn = LowerTypeReturn  #:

LowerRectTypeInput = Union[tf.Tensor, np.ndarray, Iterable[float], float]  #:
LowerRectTypeReturn = Union[np.ndarray, tf.Tensor, None, bool]  #:

UpperRectTypeInput = LowerTypeInput  #:
UpperRectTypeReturn = LowerTypeReturn  #:

RectLowerReturnType = Union[np.ndarray, tf.Tensor, float]  #:
RectUpperReturnType = RectLowerReturnType  #:
RectLimitsReturnType = Tuple[RectLowerReturnType, RectUpperReturnType]  #:
RectLimitsTFReturnType = Tuple[tf.Tensor, tf.Tensor]  #:
RectLimitsNPReturnType = Tuple[np.ndarray, np.ndarray]  #:

RectLimitsInputType = Union[LowerRectTypeInput, UpperRectTypeInput]  #:

LimitsType = Union[Tuple[Tuple[float, ...]], Tuple[float, ...], bool, "zfit.Space"]  #:
LimitsTypeSimpleInput = Union[Tuple[float, float], bool]  #:
LimitsTypeInput = Union[Tuple[Tuple[Tuple[float, ...]]], Tuple[float, float], bool]  #:
LimitsTypeReturn = Union[
    Tuple[Tuple[Tuple[float, ...]], Tuple[Tuple[float, ...]]], None, bool
]  #:

LimitsTypeInput = Union[
    "zfit.core.interfaces.ZfitLimit", RectLimitsInputType, bool, None
]  #:
LimitsFuncTypeInput = Union[LimitsTypeInput, Callable]  #:
LimitsTypeReturn = Union[Tuple[np.ndarray, np.ndarray], None, bool]  #:

_IterLimitsTypeReturn = Union[
    Tuple["zfit.Space"], Tuple[Tuple[Tuple[float]]], Tuple[Tuple[float]]
]  #:

AxesTypeInput = Union[int, Iterable[int]]  #:
AxesTypeReturn = Union[Tuple[int], None]  #:

ObsTypeInput = Union[str, Iterable[str], "zfit.Space"]  #:
ObsTypeReturn = Union[Tuple[str, ...], None]  #:
ObsType = Tuple[str]  #:

# Space
SpaceOrSpacesTypeInput = Union["zfit.Space", Iterable["zfit.Space"]]  #:
SpaceType = "zfit.Space"  #:
NormInputType = Optional["zfit.Space"]  #:

# Data
XType = Union[float, tf.Tensor]  #:
XTypeInput = Union[np.ndarray, tf.Tensor, "zfit.Data"]  #:
XTypeReturnNoData = Union[np.ndarray, tf.Tensor]  #:
XTypeReturn = Union[XTypeReturnNoData, "zfit.Data"]  #:
NumericalTypeReturn = Union[tf.Tensor, np.array]  #:

DataInputType = Union["zfit.Data", Iterable["zfit.Data"]]  #:
BinnedDataInputType = Union[PlottableHistogram, Iterable[PlottableHistogram]]  #:
ZfitBinnedDataInputType = Union[
    "zfit.data.BinnedData", Iterable["zfit.data.BinnedData"]
]  #:
AnyDataInputType = Union[DataInputType, BinnedDataInputType]  #:

WeightsStrInputType = Union[tf.Tensor, None, np.ndarray, str]  #:
WeightsInputType = Union[tf.Tensor, None, np.ndarray]  #:

# Models
ModelsInputType = Union[
    "zfit.core.interfaces.ZfitModel", Iterable["zfit.core.interfaces.ZfitModel"]
]  #:

PDFInputType = Union[
    "zfit.core.interfaces.ZfitPDF", Iterable["zfit.core.interfaces.ZfitPDF"]
]  #:
BinnedPDFInputType = Union[
    "zfit.core.interfaces.ZfitBinnedPDF", Iterable["zfit.core.interfaces.ZfitBinnedPDF"]
]  #:

FuncInputType = Union[
    "zfit.core.interfaces.ZfitFunc", Iterable["zfit.core.interfaces.ZfitFunc"]
]  #:

NumericalScalarType = Union[
    int, float, complex, tf.Tensor, "zfit.core.interfaces.ZfitParameter"
]  #:
NumericalType = Union[
    int, float, np.ndarray, tf.Tensor, "zfit.core.interfaces.ZfitParameter"
]  #:

# Integer sampling
nSamplingTypeIn = Union[int, tf.Tensor, str]  #:

ConstraintsTypeInput = Optional[
    Union[
        Iterable[Union["zfit.core.interfaces.ZfitConstraint", Callable]],
        "zfit.core.interfaces.ZfitConstraint",
        Callable,
    ]
]  #:

# Parameter
ParamsTypeOpt = Optional[Iterable["zfit.core.interfaces.ZfitParameter"]]  #:
ParamsNameOpt = Optional[Union[str, List[str]]]  #:
ParamsOrNameType = Optional[Union[ParamsTypeOpt, Iterable[str]]]  #:
ParameterType = TypeVar(
    "ParameterType", bound=Dict[str, "zfit.core.interfaces.ZfitParameter"]
)  #:
ParametersType = Iterable[ParameterType]
ParamTypeInput = TypeVar(
    "ParamTypeInput", "zfit.core.interfaces.ZfitParameter", NumericalScalarType
)  #:

ExtendedInputType = Optional[Union[bool, ParamTypeInput]]  #:

# Zfit Structure
BaseObjectType = Union[
    "zfit.core.interfaces.ZfitParameter",
    "zfit.core.interfaces.ZfitFunction",
    "zfit.core.interfaces.ZfitPDF",
]  #:
DependentsType = OrderedSet  #:

# Caching
CacherOrCachersType = Union[
    "zfit.core.interfaces.ZfitGraphCachable",
    Iterable["zfit.core.interfaces.ZfitGraphCachable"],
]  #:

OrderedDict = Dict

LimitsDictAxes = Dict[Tuple[int], "zfit.core.interfaces.ZfitLimit"]  #:
LimitsDictObs = Dict[Tuple[str], "zfit.core.interfaces.ZfitLimit"]  #:
LimitsDictNoCoords = Union[LimitsDictAxes, LimitsDictObs]  #:
LimitsDictWithCoords = Dict[str, LimitsDictNoCoords]  #:

BinningInput = Union[Iterable["ZfitBinning"], "ZfitBinning", int]  #:
OptionsInputType = Optional[Mapping[str, object]]  #:
ConstraintsInputType = Optional[
    Union[
        "zfit.core.interfaces.ZfitConstraint",
        Iterable["zfit.core.interfaces.ZfitConstraint"],
    ]
]  #:
ArrayLike = tf.types.experimental.TensorLike  #:



#  Copyright (c) 2023 zfit

import pytest

try:
    import ROOT
except ImportError:
    pytest.skip("ROOT not installed", allow_module_level=True)

import zfit
from zfit.loss import UnbinnedNLL
from zfit.minimize import Minuit


@pytest.fixture
def toy_data():
    # Generate toy data with RooFit
    x = ROOT.RooRealVar("x", "Mass", 0, 10)
    mean = ROOT.RooRealVar("mean", "mean", 5, 0, 10)
    sigma = ROOT.RooRealVar("sigma", "sigma", 1, 0, 5)
    fraction = ROOT.RooRealVar("fraction", "fraction", 0.5, 0, 1)
    gauss = ROOT.RooGaussian("gauss", "Gaussian", x, mean, sigma)
    expo = ROOT.RooExponential("expo", "Exponential", x, ROOT.RooFit.RooConst(-0.3))
    model = ROOT.RooAddPdf(
        "model", "Model", ROOT.RooArgList(gauss, expo), ROOT.RooArgList(fraction)
    )
    data = model.generate(ROOT.RooArgSet(x), 100000)
    return data


def test_roofit_vs_zfit(toy_data):
    # Perform fit with RooFit
    x = ROOT.RooRealVar("x", "Mass", 0, 10)
    mean = ROOT.RooRealVar("mean", "mean", 5, 0, 10)
    sigma = ROOT.RooRealVar("sigma", "sigma", 1, 0, 5)
    fraction = ROOT.RooRealVar("fraction", "fraction", 0.5, 0, 1)
    gauss = ROOT.RooGaussian("gauss", "Gaussian", x, mean, sigma)
    expo = ROOT.RooExponential("expo", "Exponential", x, ROOT.RooFit.RooConst(-0.3))
    model = ROOT.RooAddPdf(
        "model", "Model", ROOT.RooArgList(gauss, expo), ROOT.RooArgList(fraction)
    )
    roofit_result = model.fitTo(toy_data, ROOT.RooFit.Save(True), ROOT.RooFit.Hesse())
    # run hesse

    # Perform fit with zfit
    obs = zfit.Space("x", limits=(0, 10))
    mean = zfit.Parameter("mean", 5, 0, 10)
    sigma = zfit.Parameter("sigma", 1, 0, 5, step_size=0.01)
    fraction = zfit.Parameter("fraction", 0.5, 0, 1)
    gauss = zfit.pdf.Gauss(obs=obs, mu=mean, sigma=sigma)
    expo = zfit.pdf.Exponential(obs=obs, lambda_=zfit.Parameter("lambda", -0.3))
    model = zfit.pdf.SumPDF([gauss, expo], fracs=fraction)
    toy_data_zfit = zfit.data.Data.from_numpy(obs=obs, array=toy_data.to_numpy()["x"])
    nll = UnbinnedNLL(model=model, data=toy_data_zfit, options={"subtr_const": False})
    minimizer = Minuit()

    zfit_result = minimizer.minimize(nll)
    zfit_result.hesse()

    # Compare uncertainties
    mean_unc = zfit_result.params[mean]["hesse"]["error"]
    assert (
        pytest.approx(roofit_result.floatParsFinal().find("mean").getError(), rel=0.05)
        == mean_unc
    )
    sigma_unc = zfit_result.params[sigma]["hesse"]["error"]
    assert (
        pytest.approx(roofit_result.floatParsFinal().find("sigma").getError(), rel=0.15)
        == sigma_unc
    )
    # Compare results
    assert pytest.approx(roofit_result.minNll(), rel=1e-3) == zfit_result.fmin
    assert (
        pytest.approx(
            roofit_result.floatParsFinal().find("mean").getVal(), abs=3 * mean_unc
        )
        == mean.value()
    )
    assert (
        pytest.approx(
            roofit_result.floatParsFinal().find("sigma").getVal(), abs=3 * sigma_unc
        )
        == sigma.value()
    )



#  Copyright (c) 2023 zfit

import numpy as np
import tensorflow as tf
from dotmap import DotMap

from .util.execution import RunManager

run = RunManager()


def set_seed(seed=None, numpy=None, backend=None):
    """Set random seed for zfit, numpy and the backend.

    Args:
        seed (int, optional): Seed to set for the random number generator within zfit. If `None` (default), the seed is set to a random value.
        numpy (int, bool, optional): Seed to set for numpy. If `True` (default), the same seed as for zfit is used. If `False`, the seed is not set.
        backend (int, bool, optional): Seed to set for the backend. If `True` (default), the same seed as for zfit is used. If `False`, the seed is not set.
    """
    if seed is None:
        import os

        random_data = os.urandom(4)
        backend_seed = int.from_bytes(random_data, byteorder="big")
        backend_seed = int(
            abs(backend_seed) % 2**31
        )  # make sure it's positive and not too large
        seed = backend_seed

    if numpy is None:
        numpy = True
    if backend is None:
        backend = True

    if numpy is True:
        numpy = seed
    if backend is True:
        backend = seed

    if numpy is not None and numpy is not False:
        np.random.seed(numpy)
    if backend is not None and backend is not False:
        tf.random.set_seed(backend)

    from .z.random import get_prng

    get_prng().reset_from_seed(seed)


_verbosity = 0


def set_verbosity(verbosity):
    global _verbosity
    _verbosity = verbosity


def get_verbosity():
    return _verbosity


ztypes = DotMap(
    {
        "float": tf.float64,
        "complex": tf.complex128,
        "int": tf.int64,
        "auto_upcast": True,
    }
)
upcast_ztypes = {
    tf.float16: tf.float64,
    tf.float32: tf.float64,
    tf.float64: tf.float64,
    tf.complex64: tf.complex128,
    tf.complex128: tf.complex128,
    tf.int8: tf.int64,
    tf.int16: tf.int64,
    tf.int32: tf.int64,
    tf.int64: tf.int64,
}

options = DotMap(
    {
        "epsilon": 1e-8,
        "numerical_grad": None,
        "advanced_warning": True,
        "changed_warning": True,
    }
)

advanced_warnings = DotMap(
    {
        "sum_extended_frac": True,
        "exp_shift": True,
        "py_func_autograd": True,
        "inconsistent_fitrange": True,
        "extended_in_UnbinnedNLL": True,
        "all": True,
    }
)

changed_warnings = DotMap(
    {
        "new_sum": True,
        "all": True,
    }
)



#  Copyright (c) 2023 zfit
"""Baseclass for ``Function``. Inherits from Model.

TODO(Mayou36): subclassing?
"""
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import zfit

import abc
import typing

from ..util.exception import ShapeIncompatibleError, SpecificFunctionNotImplemented
from .basemodel import BaseModel
from .interfaces import ZfitFunc
from ..settings import ztypes
from ..util import ztyping


class BaseFuncV1(BaseModel, ZfitFunc):
    def __init__(
        self,
        obs=None,
        dtype: type = ztypes.float,
        name: str = "BaseFunc",
        params: typing.Any = None,
    ):
        """TODO(docs): explain subclassing."""
        super().__init__(obs=obs, dtype=dtype, name=name, params=params)

    def _func_to_integrate(self, x: ztyping.XType):
        return self.func(x=x)

    def _func_to_sample_from(self, x):
        return self.func(x=x)

    # TODO(Mayou36): how to deal with copy properly?
    def copy(self, **override_params):
        new_params = self.params
        new_params.update(override_params)
        return type(self)(new_params)

    def gradient(
        self,
        x: ztyping.XType,
        norm: ztyping.LimitsType = None,
        params: ztyping.ParamsTypeOpt = None,
    ):
        # TODO(Mayou36): well, really needed... this gradient?
        raise NotImplementedError("What do you need? Use tf.gradient...")

    @abc.abstractmethod
    def _func(self, x):
        raise SpecificFunctionNotImplemented

    def func(self, x: ztyping.XType, name: str = "value") -> ztyping.XType:
        """The function evaluated at ``x``.

        Args:
            x:
            name:

        Returns:
             # TODO(Mayou36): or dataset? Update: rather not, what would obs be?
        """
        with self._convert_sort_x(x) as x:
            return self._single_hook_value(x=x, name=name)

    def _single_hook_value(self, x, name):
        return self._hook_value(x, name)

    def _hook_value(self, x, name="_hook_value"):
        return self._call_value(x=x, name=name)

    def _call_value(self, x, name):
        try:
            return self._func(x=x)
        except ValueError as error:
            raise ShapeIncompatibleError(
                "Most probably, the number of obs the func was designed for"
                "does not coincide with the `n_obs` from the `space`/`obs`"
                "it received on initialization."
            ) from error

    def as_pdf(self) -> zfit.core.interfaces.ZfitPDF:
        """Create a PDF out of the function.

        Returns:
            A PDF with the current function as the unnormalized probability.
        """
        from zfit.core.operations import convert_func_to_pdf

        return convert_func_to_pdf(func=self)

    def _check_input_norm_range_default(
        self, norm_range, caller_name="", none_is_error=True
    ):  # TODO(Mayou36): default
        return self._check_input_norm(norm=norm_range, none_is_error=none_is_error)



#  Copyright (c) 2023 zfit
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Code was copied directly from tf_quant_finance.math.root_search.brent.
# The repository is deprecated and will be removed in the future.
"""Common data types."""

from typing import TypeVar
import numpy as np
import tensorflow as tf

__all__ = ["BoolTensor", "IntTensor", "RealTensor", "ComplexTensor", "StringTensor"]

tensor_like = (np.ndarray, tf.Tensor, tf.TensorSpec)

# A type that represents a boolean `Tensor`
BoolTensor = TypeVar("BoolTensor", *tensor_like)

# A type that represents int32 or int64 `Tensor`s
IntTensor = TypeVar("IntTensor", *tensor_like)

# A type that represents float or double `Tensor`s
RealTensor = TypeVar("RealTensor", *tensor_like)

# A type that represents a float `Tensor`
FloatTensor = TypeVar("FloatTensor", *tensor_like)

# A type that represents a double `Tensor`
DoubleTensor = TypeVar("DoubleTensor", *tensor_like)

# 'A type that represents complex64 or complex128 `Tensor`s
ComplexTensor = TypeVar("ComplexTensor", *tensor_like)

# A type that represents a string `Tensor`
StringTensor = TypeVar("StringTensor", *tensor_like)
